import os
import math
import logging
import pathlib

import numpy as np
import pandas as pd
import opendssdirect as dss

from .pydss_parameters import *
from jade.utils.timing_utils import track_timing, Timer

from disco import timer_stats_collector
from disco.enums import LoadMultiplierType
from disco.exceptions import (
    OpenDssCompileError,
    OpenDssConvergenceError,
    UpgradesExternalCatalogRequired,
    UpgradesExternalCatalogMissingObjectDefinition,
    InvalidOpenDssElementError,
)
from disco.models.upgrade_cost_analysis_generic_input_model import (
    _extract_specific_model_properties_, 
    LineCodeCatalogModel, LineGeometryCatalogModel,
    LineModel, LineCatalogModel,
    TransformerModel, TransformerCatalogModel,
)
from disco.models.upgrade_cost_analysis_generic_output_model import UpgradesCostResultSummaryModel, \
    CapacitorControllerResultType,  VoltageRegulatorResultType, EquipmentUpgradeStatusModel

logger = logging.getLogger(__name__)

DSS_XFMR_FLOAT_FIELDS = _extract_specific_model_properties_(model_name=TransformerModel, field_type_key="type", field_type_value="number")
DSS_XFMR_INT_FIELDS = _extract_specific_model_properties_(model_name=TransformerModel, field_type_key="type", field_type_value="integer")
DSS_LINE_FLOAT_FIELDS =  _extract_specific_model_properties_(model_name=LineModel, field_type_key="type", field_type_value="number")
DSS_LINE_INT_FIELDS = _extract_specific_model_properties_(model_name=LineModel, field_type_key="type", field_type_value="integer")
DSS_LINECODE_FLOAT_FIELDS =  _extract_specific_model_properties_(model_name=LineCodeCatalogModel, field_type_key="type", field_type_value="number")
DSS_LINECODE_INT_FIELDS =  _extract_specific_model_properties_(model_name=LineCodeCatalogModel, field_type_key="type", field_type_value="integer")
DSS_LINEGEOMETRY_FLOAT_FIELDS =  _extract_specific_model_properties_(model_name=LineGeometryCatalogModel, field_type_key="type", field_type_value="number")
DSS_LINEGEOMETRY_INT_FIELDS =  _extract_specific_model_properties_(model_name=LineGeometryCatalogModel, field_type_key="type", field_type_value="integer")
DSS_UNIT_CONFIG = {1: "mi", 2: "kft", 3: "m", 4: "Ft", 5: "in", 6: "cm",
                    0: "none"  # 0 maps to none, which means impedance units and line length units match
                    }

@track_timing(timer_stats_collector)
def reload_dss_circuit(dss_file_list, commands_list=None,  **kwargs):
    """This function clears the circuit and loads dss files and commands.
    Also solves the circuit and checks for convergence errors

    Parameters
    ----------
    dss_file_list
    commands_list

    Returns
    -------

    """
    logger.info("Reloading OpenDSS circuit")
    check_dss_run_command("clear")
    if dss_file_list is None:
        raise Exception("No OpenDSS files have been passed to be loaded.")
    for dss_file in dss_file_list:
        logger.info(f"Redirecting '{dss_file}'.")
        check_dss_run_command(f"Redirect '{dss_file}'")
    dc_ac_ratio = kwargs.get('dc_ac_ratio', None)
    if dc_ac_ratio is not None:
        change_pv_pctpmpp(dc_ac_ratio=dc_ac_ratio)
    if commands_list is not None:
        logger.info(f"Running {len(commands_list)} dss commands")
        for command_string in commands_list:
            check_dss_run_command(command_string)
            if "new " in command_string.lower():
                check_dss_run_command("CalcVoltageBases")
    enable_pydss_solve = kwargs.get("enable_pydss_solve", False)
    raise_exception = kwargs.get("raise_exception", True)
    if enable_pydss_solve:
        pydss_params = define_initial_pydss_settings(**kwargs)
        circuit_solve_and_check(raise_exception=raise_exception, **pydss_params)
        return pydss_params
    else:
        max_control_iterations = kwargs.get("max_control_iterations", None)
        if max_control_iterations is not None:
            dss.Solution.MaxControlIterations(max_control_iterations)
        circuit_solve_and_check(raise_exception=raise_exception)
        return kwargs


def run_selective_master_dss(master_filepath, **kwargs):
    """This function executes master.dss file line by line and ignores some commands that Solve yearly mode,
    export or plot data.

    Parameters
    ----------
    master_filepath

    Returns
    -------

    """
    run_dir = os.getcwd()
    check_dss_run_command("Clear")
    # logger.info("Redirecting master file:")
    # check_dss_run_command(f"Redirect {master_filepath}")

    # do this instead of redirect master to ignore some lines (e.g., that solve for the whole year)
    os.chdir(os.path.dirname(master_filepath))
    logger.debug(master_filepath)
    with open(master_filepath, "r") as fr:
        tlines = fr.readlines()
    for line in tlines:
        if ('Solve'.lower() in line.lower()) or ('Export'.lower() in line.lower()) or ('Plot'.lower() in line.lower()):
            logger.info(f"Skipping this line: {line}")
            continue
        else:
            check_dss_run_command(f"{line}")
    circuit_solve_and_check(raise_exception=True, **kwargs)
    os.chdir(run_dir)
    return


@track_timing(timer_stats_collector)
def circuit_solve_and_check(raise_exception=False, **kwargs):
    """This function solves the circuit (both OpenDSS and PyDSS-if enabled)
    and can raise exception if convergence error occurs

    Parameters
    ----------
    raise_exception
    kwargs

    Returns
    -------

    """
    calcvoltagebases = kwargs.pop("calcvoltagebases", False)
    if calcvoltagebases:
        check_dss_run_command("CalcVoltageBases")
    dss_pass_flag = dss_solve_and_check(raise_exception=raise_exception)
    pass_flag = dss_pass_flag
    enable_pydss_solve = kwargs.get("enable_pydss_solve", False)
    if enable_pydss_solve:  # if pydss solver is also to be used
        pydss_pass_flag = pydss_solve_and_check(raise_exception=raise_exception, **kwargs)
        pass_flag = dss_pass_flag and pydss_pass_flag
    return pass_flag


def dss_solve_and_check(raise_exception=False):
    """This function solves OpenDSS and returns bool flag which shows if it has converged or not.

    Parameters
    ----------
    raise_exception

    Returns
    -------
    bool
    """
    dss.Solution.Solve()
    logger.debug("Solving circuit using OpenDSS")
    # check_dss_run_command('CalcVoltageBases')
    dss_pass_flag = dss.Solution.Converged()
    if not dss_pass_flag:
        logger.info(f"OpenDSS Convergence Error")
        if raise_exception:
            raise OpenDssConvergenceError("OpenDSS solution did not converge")
    return dss_pass_flag


def dss_run_command_list(command_list):
    for command_string in command_list:
        check_dss_run_command(command_string)
    return


def write_text_file(string_list, text_file_path, **kwargs):
    """This function writes the string contents of a list to a text file

    Parameters
    ----------
    string_list
    text_file_path

    Returns
    -------

    """
    num_new_lines = kwargs.get("num_new_lines", 2)
    breaks = "\n"*num_new_lines
    pathlib.Path(text_file_path).write_text(breaks.join(string_list))


def create_upgraded_master_dss(dss_file_list, upgraded_master_dss_filepath, original_master_filename="master.dss"):
    """Function to create master dss with redirects to upgrades dss file.
    The redirect paths in this file are:
    * absolute path - to the original master dss file
    * relative path (relative to the upgraded_master dss file) if upgrades dss file"""
    command_list = []
    for filename in dss_file_list:
        if os.path.basename(filename) == original_master_filename:
            new_filename = _get_master_dss_filepath(filename, upgraded_master_dss_filepath)
        else:    
            new_filename = os.path.relpath(filename, os.path.dirname(upgraded_master_dss_filepath))
        command_list.append(f"Redirect {new_filename}")
    return command_list


def _get_master_dss_filepath(original_master, upgraded_master):
    if os.path.isabs(upgraded_master):
        # Here it is not possible to use a relative path in all cases.
        # The runtime output directory may have a different root than the source files.
        return os.path.abspath(original_master)

    return os.path.relpath(original_master, os.path.dirname(upgraded_master))


def create_dataframe_from_nested_dict(user_dict, index_names):
    """This function creates dataframe from a nested dictionary

    Parameters
    ----------
    user_dict
    index_names

    Returns
    -------
    DataFrame
    """
    df = pd.DataFrame.from_dict({(i, j): user_dict[i][j]
                                 for i in user_dict.keys()
                                 for j in user_dict[i].keys()},
                                orient='index')
    df.index.names = index_names
    return df.reset_index()


def get_dictionary_of_duplicates(df, subset, index_field):
    """This creates a mapping dictionary of duplicate indices in a dataframe

    Parameters
    ----------
    df
    subset
    index_field

    Returns
    -------
    Dictionary
    """
    df.set_index(index_field, inplace=True)
    df = df[df.duplicated(keep=False, subset=subset)]
    tuple_list = df.groupby(subset).apply(lambda x: tuple(x.index)).tolist()
    mapping_dict = {v: tup[0] for tup in tuple_list for v in tup}
    return mapping_dict


def convert_length_units(length, unit_in, unit_out):
    """Length unit converter"""
    LENGTH_CONVERSION = {'mm': 0.001, 'cm': 0.01, 'm': 1.0, 'km': 1000., "mi": 1609.34, "kft": 304.8, 
          "ft": 0.3048,  "in": 0.0254,}
    return length*LENGTH_CONVERSION[unit_in]/LENGTH_CONVERSION[unit_out]


def get_scenario_name(enable_pydss_solve, pydss_volt_var_model):
    """This function determines the controller scenario 

    Parameters
    ----------
    enable_pydss_solve : bool
    pydss_volt_var_model 

    Returns
    -------
    str
    """
    if enable_pydss_solve:
        # scenario = pydss_volt_var_model.control1  # TODO can read in name instead
        scenario = "control_mode"
    else:
        scenario = "pf1"
    return scenario


def convert_dict_nan_to_none(temp):
    """Convert np.nan in dictionary to None.
    This does change the data type of the field."""
    for key, value in temp.items():
        if isinstance(value, dict):
            df = pd.DataFrame([value])
            if df.isna().values.any():
                df = df.astype(object).where(df.notna(), None)  # replace NaN with None
                value = df.to_dict() 
            else:
                continue
        elif isinstance(value, list) and bool(value) and isinstance(value[0], dict):  # list of dicts
            df = pd.DataFrame(value)
            if df.isna().values.any():
                df = df.astype(object).where(df.notna(), None)  # replace NaN with None
                value = df.to_dict(orient="records")          
            else:
                continue  
        temp[key] = value
    return temp


@track_timing(timer_stats_collector)
def change_pv_pctpmpp(dc_ac_ratio):
    """This function changes PV system pctpmpp based on passed dc-ac ratio
    newpctpmpp = oldpctpmpp / dc_ac_ratio
    """
    dss.PVsystems.First()
    for i in range(dss.PVsystems.Count()):
        newpctpmpp = int(dss.Properties.Value('%Pmpp')) / dc_ac_ratio
        command_string = f"Edit PVSystem.{dss.PVsystems.Name()} %Pmpp={newpctpmpp}"
        check_dss_run_command(command_string)
        dss.PVsystems.Next()


def get_feeder_stats(dss):
    """This function gives metadata stats for a feeder 

    Parameters
    ----------
    dss

    Returns
    -------
    dict
    """
    load_kw = 0
    load_kVABase = 0
    pv_kw = 0
    pv_kVARated = 0
    
    load_df = dss.utils.loads_to_dataframe()
    if len(load_df) > 0:
        load_kw = load_df['kW'].sum()
        load_kVABase = load_df['kVABase'].sum()
    if dss.PVsystems.Count() > 0:
        pv_df = dss.utils.pvsystems_to_dataframe()
        pv_kw = pv_df['kW'].sum()
        pv_kVARated = pv_df['kVARated'].sum()
    data_dict = {
    'total_load(kVABase)': load_kVABase,
    'total_load(kW)': load_kw,
    'total_PV(kW)': pv_kw,
    'total_PV(kVARated)': pv_kVARated,
    }
    return data_dict


def get_upgrade_stage_stats(dss, upgrade_stage, upgrade_type, xfmr_loading_df, line_loading_df, bus_voltages_df, **kwargs):
    """This function gives upgrade stage stats for a feeder 
    upgrade_stage can be Initial or Final
    upgrade_type can be thermal or voltage
    
    """
    final_dict = {"stage": upgrade_stage, "upgrade_type": upgrade_type}
    ckt_info_dict = get_circuit_info()
    final_dict["feeder_components"] = ckt_info_dict
    final_dict["feeder_components"].update({
                                        "num_nodes": dss.Circuit.NumNodes(),
                                        "num_loads": dss.Loads.Count(),
                                        "num_lines": dss.Lines.Count(),
                                        "num_transformers": dss.Transformers.Count(),
                                        "num_pv_systems": dss.PVsystems.Count(),
                                        "num_capacitors": dss.Capacitors.Count(),
                                        "num_regulators": dss.RegControls.Count(),
                                        } )
    equipment_dict = combine_equipment_health_stats(xfmr_loading_df, line_loading_df, bus_voltages_df, **kwargs)
    final_dict.update(equipment_dict)
    return final_dict


def combine_equipment_health_stats(xfmr_loading_df, line_loading_df, bus_voltages_df, **kwargs):
    line_properties = kwargs.get("line_properties", 
                                 ['name', 'phases','normamps', 'kV', 'line_placement',  'length', 'units', 'max_amp_loading', 
                                  'max_per_unit_loading', 'status'])
    xfmr_properties = kwargs.get("xfmr_properties", 
                                 ['name', 'phases', 'windings', 'conns', 'kV', 'kVA', 'amp_limit_per_phase','max_amp_loading', 
                                  'max_per_unit_loading', 'status']  )
    voltage_properties = kwargs.get("voltage_properties", 
                                    ['name', 'max_per_unit_voltage', 'min_per_unit_voltage',  'overvoltage_violation', 
                                     'max_voltage_deviation', 'undervoltage_violation', 'min_voltage_deviation'])
    capacitors_df = kwargs.get("capacitors_df", pd.DataFrame())
    regcontrols_df = kwargs.get("regcontrols_df", pd.DataFrame())
    capacitor_properties = kwargs.get("capacitor_properties", 
                                      ['capacitor_name','capcontrol_present',  'capcontrol_type', 'capcontrol_name', 'kv', 'kvar',
                                       'phases', 'DeadTime', 'Delay', 'OFFsetting', 'ONsetting'])
    regcontrol_properties = kwargs.get("regcontrol_properties", 
                                       ['name', 'transformer', 'vreg', 'band', 'ptratio', 'delay', 'at_substation_xfmr_flag'])
    
    final_dict = {}
    # some file reformatting
    if "windings" in xfmr_properties:
        xfmr_loading_df["windings"] = xfmr_loading_df["windings"].astype(int)
    final_dict.update({"transformer": xfmr_loading_df[xfmr_properties].to_dict(orient="records")})
    final_dict.update({"line": line_loading_df[line_properties].to_dict(orient="records")})
    final_dict.update({"bus_voltage": bus_voltages_df[voltage_properties].to_dict(orient="records")})
    if not capacitors_df.empty :
        final_dict.update({"capacitor_control": capacitors_df[capacitor_properties].to_dict(orient="records")})
    else :
        final_dict.update({"capacitor_control": []})
    if not regcontrols_df.empty:
        final_dict.update({"regulator_control": regcontrols_df[regcontrol_properties].to_dict(orient="records")})
    else:
        final_dict.update({"regulator_control": []})
    return final_dict


def get_circuit_info():
    """This collects circuit information: source bus, feeder head info, substation xfmr information

    Returns
    -------
    Dictionary
    """
    data_dict = {}
    dss.Vsources.First()
    data_dict['source_bus'] = dss.CktElement.BusNames()[0].split(".")[0]
    data_dict["feeder_head_name"] = dss.Circuit.Name()
    dss.Circuit.SetActiveBus(data_dict['source_bus'])
    data_dict["feeder_head_basekv"] = dss.Bus.kVBase()
    data_dict["source_num_nodes"] = dss.Bus.NumNodes()
    data_dict["total_num_buses_in_circuit"] = len(dss.Circuit.AllBusNames())
    if data_dict["source_num_nodes"] > 1:
        data_dict["feeder_head_basekv"] = round(data_dict["feeder_head_basekv"] * math.sqrt(3), 1)
    data_dict["substation_xfmr"] = None

    all_xfmr_df = get_thermal_equipment_info(compute_loading=False, equipment_type="transformer")
    all_xfmr_df["substation_xfmr_flag"] = all_xfmr_df.apply(lambda x: int(
        data_dict["source_bus"].lower() in x['bus_names_only']), axis=1)
    if len(all_xfmr_df.loc[all_xfmr_df["substation_xfmr_flag"] == True]) > 0:
        data_dict["substation_xfmr"] = all_xfmr_df.loc[all_xfmr_df["substation_xfmr_flag"] ==
                                                       True].to_dict(orient='records')[0]
        # this checks if the voltage kVs are the same for the substation transformer
        data_dict["substation_xfmr"]["is_autotransformer_flag"] = len(set(data_dict["substation_xfmr"]["kVs"])) <= 1
    return data_dict


def summarize_upgrades_outputs(overall_outputs, **kwargs):
    """This function creates summary of upgrades and costs results"""
    summary = {"results": {}}
    summary["results"]["name"] = kwargs.get("job_name", None)
    violation_summary = pd.DataFrame(overall_outputs["violation_summary"])
    thermal_violations = sum(violation_summary.loc[(violation_summary["stage"] == "final") & (violation_summary["upgrade_type"] == "thermal")][["num_line_violations", "num_transformer_violations"]].sum())
    voltage_violations = sum(violation_summary.loc[(violation_summary["stage"] == "final") & (violation_summary["upgrade_type"] == "voltage")][["num_voltage_violation_buses"]].sum())
    summary["results"]["num_violations"] = thermal_violations + voltage_violations
    if overall_outputs["costs_per_equipment"]:
        summary["results"]["total_cost_usd"] = pd.DataFrame(overall_outputs["costs_per_equipment"])["total_cost_usd"].sum()
    else:
        summary["results"]["total_cost_usd"] = 0
    return summary


def create_thermal_output_summary(all_original_equipment, all_latest_equipment, thermal_equipment_type_list,
                                    props_dict, thermal_cost_df, upgrades_dict, output_cols):
    """This function creates the thermal output summary file"""
    new_thermal_df = pd.DataFrame(columns=output_cols)
    for equipment_type in thermal_equipment_type_list:
        latest_equipment_df = pd.DataFrame(all_latest_equipment[equipment_type])
        latest_equipment_df = latest_equipment_df.rename(columns={props_dict[equipment_type]["identifier"]: "equipment_name"})
        original_equipment_df = pd.DataFrame(all_original_equipment[equipment_type])
        original_equipment_df = original_equipment_df.rename(columns={props_dict[equipment_type]["identifier"]: "equipment_name"})
        temp_upgrade_df = upgrades_dict[equipment_type]
        temp_upgrade_df = temp_upgrade_df.rename(columns={"final_equipment_name": "equipment_name"})
        if (temp_upgrade_df.empty) and (original_equipment_df.empty):  # if there are no equipment of that type
            continue 
        new_df = latest_equipment_df.copy(deep=True)
        new_df = pd.concat([new_df, pd.DataFrame(columns=list(set(output_cols)-set(new_df.columns)))], axis=1)
        new_df["equipment_type"] = equipment_type
        new_df["total_cost_usd"] = 0
        new_df["status"] = EquipmentUpgradeStatusModel.unchanged.value
        
        if not temp_upgrade_df.empty:
            temp_cost_df = thermal_cost_df.loc[thermal_cost_df.type.str.lower() == equipment_type.lower()]
            temp_cost_df = temp_cost_df.rename(columns={"final_equipment_name": "equipment_name"})
            replaced = list(temp_upgrade_df.loc[temp_upgrade_df["upgrade_type"]=="upgrade"]["equipment_name"].unique())  # list of replaced equipment
            new = list(temp_upgrade_df.loc[temp_upgrade_df["upgrade_type"]=="new_parallel"]["equipment_name"].unique())  # list of new equipment
            # add upgrade status
            new_df.loc[new_df.equipment_name.isin(replaced), "status"] = EquipmentUpgradeStatusModel.replaced.value
            new_df.loc[new_df.equipment_name.isin(new), "status"] = EquipmentUpgradeStatusModel.new.value
            # add cost
            temp_cost_df.set_index("equipment_name", inplace=True)
            new_df.set_index("equipment_name", inplace=True)
            if len(temp_cost_df[temp_cost_df.index.duplicated()]) > 0:
                raise Exception("Duplicates upgrades shouldn't exist. Review logic for this feeder model case.")
            new_df.loc[new_df.index.isin(temp_cost_df.index), "total_cost_usd"] = temp_cost_df.total_cost_usd
        else:
            new = []
            replaced = []
        parameter_list = props_dict[equipment_type]["parameter_list"]
        original_equipment_df.set_index("equipment_name", inplace=True)
        for i in range(0, len(parameter_list)):
            new_df[f"parameter{i+1}_name"] = parameter_list[i]
            new_df[f"parameter{i+1}_original"] = new_df[parameter_list[i]]
            new_df[f"parameter{i+1}_upgraded"] = new_df[parameter_list[i]]
            new_df.loc[new_df.index.isin(new), f"parameter{i+1}_original"] = None # new equipment original rating is None
            new_df.loc[new_df.index.isin(replaced), f"parameter{i+1}_original"] = original_equipment_df.loc[original_equipment_df.index.isin(replaced)][parameter_list[i]]
    
        new_df.reset_index(inplace=True)
        new_df = new_df[output_cols]
        new_thermal_df = pd.concat([new_thermal_df, new_df])
    new_thermal_df = new_thermal_df.replace({np.NaN: None})
    return new_thermal_df


def create_capacitor_output_summary(temp_upgrade_df, temp_cost_df, latest_equipment_df, output_cols, equipment_type):
    # create new dataframe
    new_df = latest_equipment_df.copy(deep=True)
    new_df = pd.concat([new_df, pd.DataFrame(columns=list(set(output_cols)-set(new_df.columns)))], axis=1)
    new_df["equipment_type"] = equipment_type
    new_df["total_cost_usd"] = 0
    new_df["status"] = EquipmentUpgradeStatusModel.unchanged.value
    
    if temp_upgrade_df.empty:  # if there are no upgrades of this equipment type
        return new_df, [], []
    
    new = list(temp_upgrade_df.loc[temp_upgrade_df["new_controller_added"]]["equipment_name"].unique())  # list of new equipment
    setting_changed = list(temp_upgrade_df.loc[temp_upgrade_df["controller_settings_modified"]]["equipment_name"].unique())  # list of setting_changed equipment
    # get unit cost
    unit_cost_calc = temp_cost_df.loc[temp_cost_df.type == CapacitorControllerResultType.change_cap_control.value]
    if not unit_cost_calc.empty:
        setting_changed_unit_cost = (unit_cost_calc["total_cost_usd"] / unit_cost_calc["count"]).values[0]
    else: 
        setting_changed_unit_cost = 0
    unit_cost_calc = temp_cost_df.loc[temp_cost_df.type == CapacitorControllerResultType.add_new_cap_controller.value]
    if not unit_cost_calc.empty:
        add_new_unit_cost = (unit_cost_calc["total_cost_usd"] / unit_cost_calc["count"]).values[0]
    else:
        add_new_unit_cost = 0
    # add upgrade status and cost
    new_df.loc[new_df.equipment_name.isin(setting_changed), "status"] = EquipmentUpgradeStatusModel.setting_changed.value
    new_df.loc[new_df.equipment_name.isin(setting_changed), "total_cost_usd"] = setting_changed_unit_cost
    new_df.loc[new_df.equipment_name.isin(new), "status"] = EquipmentUpgradeStatusModel.new.value
    new_df.loc[new_df.equipment_name.isin(new), "total_cost_usd"] = add_new_unit_cost
    return new_df, new, setting_changed 


def create_regulator_output_summary(temp_upgrade_df, temp_cost_df, latest_equipment_df, output_cols, equipment_type):    
    # create new dataframe
    new_df = latest_equipment_df.copy(deep=True)
    new_df = pd.concat([new_df, pd.DataFrame(columns=list(set(output_cols)-set(new_df.columns)))], axis=1)
    new_df["equipment_type"] = equipment_type
    new_df["total_cost_usd"] = 0
    new_df["status"] = EquipmentUpgradeStatusModel.unchanged.value
    if temp_upgrade_df.empty:  # if there are no upgrades of this equipment type
        return new_df, [], []
    
    new = list(temp_upgrade_df.loc[temp_upgrade_df["new_controller_added"]]["equipment_name"].unique())  # list of new equipment
    setting_changed = list(temp_upgrade_df.loc[temp_upgrade_df["controller_settings_modified"]]["equipment_name"].unique())  # list of setting_changed equipment
    # get unit cost            
    unit_cost_calc = temp_cost_df.loc[temp_cost_df.type == VoltageRegulatorResultType.add_new_reg_control.value]
    if not unit_cost_calc.empty:
        new_vreg_unit_cost = (unit_cost_calc["total_cost_usd"] / unit_cost_calc["count"]).values[0]
        new_df.loc[(new_df.equipment_name.isin(new)) & (new_df.at_substation_xfmr_flag == False), "total_cost_usd"] = new_vreg_unit_cost
    unit_cost_calc = temp_cost_df.loc[temp_cost_df.type == VoltageRegulatorResultType.change_reg_control.value]
    if not unit_cost_calc.empty:
        vreg_setting_changed_unit_cost = (unit_cost_calc["total_cost_usd"] / unit_cost_calc["count"]).values[0]
        new_df.loc[(new_df.equipment_name.isin(setting_changed)) & (new_df.at_substation_xfmr_flag == False), "total_cost_usd"] = vreg_setting_changed_unit_cost
    unit_cost_calc = temp_cost_df.loc[temp_cost_df.type == VoltageRegulatorResultType.add_substation_ltc.value]
    if not unit_cost_calc.empty:
        new_ltc_unit_cost = (unit_cost_calc["total_cost_usd"] / unit_cost_calc["count"]).values[0]
        new_df.loc[(new_df.equipment_name.isin(new)) & (new_df.at_substation_xfmr_flag), "total_cost_usd"] = new_ltc_unit_cost
    unit_cost_calc = temp_cost_df.loc[temp_cost_df.type == VoltageRegulatorResultType.change_ltc_control.value]
    if not unit_cost_calc.empty:
        ltc_setting_changed_unit_cost = (unit_cost_calc["total_cost_usd"] / unit_cost_calc["count"]).values[0]
        new_df.loc[(new_df.equipment_name.isin(setting_changed)) & (new_df.at_substation_xfmr_flag), "total_cost_usd"] = ltc_setting_changed_unit_cost
    # add upgrade status
    new_df.loc[new_df.equipment_name.isin(setting_changed), "status"] = EquipmentUpgradeStatusModel.setting_changed.value
    new_df.loc[new_df.equipment_name.isin(new), "status"] = EquipmentUpgradeStatusModel.new.value
    return new_df, new, setting_changed 


def create_voltage_output_summary(all_original_equipment, all_latest_equipment, voltage_equipment_type_list,
                                    props_dict, voltage_cost_df, upgrades_dict, output_cols):
    # VOLTAGE EQUIPMENT
    voltage_upgrade_df = upgrades_dict["voltage"]
    voltage_upgrade_df = voltage_upgrade_df.rename(columns={"final_equipment_name": "equipment_name"})
    new_voltage_df = pd.DataFrame(columns=output_cols)
    
    for equipment_type in voltage_equipment_type_list:
        latest_equipment_df = pd.DataFrame(all_latest_equipment[equipment_type])
        latest_equipment_df = latest_equipment_df.rename(columns={props_dict[equipment_type]["identifier"]: "equipment_name"})
        original_equipment_df = pd.DataFrame(all_original_equipment[equipment_type])
        original_equipment_df = original_equipment_df.rename(columns={props_dict[equipment_type]["identifier"]: "equipment_name"})
        if (voltage_upgrade_df.empty) and (original_equipment_df.empty):  # if there are no equipment for voltage controls
            continue 
        if not voltage_upgrade_df.empty:  # if there are voltage upgrades, extract for this equipment type
            temp_cost_df = voltage_cost_df.loc[voltage_cost_df.type.isin(props_dict[equipment_type]["model"].list_values())]
            temp_upgrade_df = voltage_upgrade_df.loc[voltage_upgrade_df.equipment_type.str.lower() == props_dict[equipment_type]["upgrades_file_string"].lower()]
        else:
            temp_upgrade_df = pd.DataFrame(columns=voltage_upgrade_df.columns)
            temp_cost_df = pd.DataFrame(columns=voltage_cost_df.columns)
        temp_upgrade_df = temp_upgrade_df.rename(columns={"name": "equipment_name"}) 
        temp_cost_df = temp_cost_df.rename(columns={"final_equipment_name": "equipment_name"})
        if equipment_type == "capacitor_control":
            new_df, new, setting_changed = create_capacitor_output_summary(temp_upgrade_df, temp_cost_df, latest_equipment_df, output_cols, equipment_type)
        elif equipment_type == "regulator_control":
            new_df, new, setting_changed  = create_regulator_output_summary(temp_upgrade_df, temp_cost_df, latest_equipment_df, output_cols, equipment_type)
        if new_df.empty:  # if there are no equipment of this type
            continue            
        new_df.set_index("equipment_name", inplace=True)
        parameter_list = props_dict[equipment_type]["parameter_list"]
        if not original_equipment_df.empty:
            original_equipment_df.set_index("equipment_name", inplace=True)
        for i in range(0, len(parameter_list)):
            new_df[f"parameter{i+1}_name"] = parameter_list[i].lower()
            new_df[f"parameter{i+1}_original"] = new_df[parameter_list[i]]
            new_df[f"parameter{i+1}_upgraded"] = new_df[parameter_list[i]]
            new_df.loc[new_df.index.isin(new), f"parameter{i+1}_original"] = None # new equipment original rating is None
            if not original_equipment_df.empty:
                new_df.loc[new_df.index.isin(setting_changed), f"parameter{i+1}_original"] = original_equipment_df.loc[original_equipment_df.index.isin(setting_changed)][parameter_list[i]]
        new_df.reset_index(inplace=True)
        new_df = new_df[output_cols]
        new_voltage_df = pd.concat([new_voltage_df, new_df])
    return new_voltage_df


def create_overall_output_file(feeder_stats, upgrades_dict, costs_dict, **kwargs):
    """This function creates the overall output summary file
    Status can have values: unchanged, replaced, new, setting_changed
    """
    output_cols = UpgradesCostResultSummaryModel.schema(True).get("properties").keys()
    thermal_equipment_type_list = kwargs.get("thermal_equipment_type_list", ["transformer", "line"])
    voltage_equipment_type_list = kwargs.get("voltage_equipment_type_list", ["capacitor_control", "regulator_control"])
    props_dict = {"transformer": {"identifier": "name", "parameter_list": ["kVA"], },
                 "line": {"identifier": "name", "parameter_list": ["normamps"], },
                 "capacitor_control": {"identifier": "capacitor_name", "parameter_list": ["ONsetting", "OFFsetting", "Delay"], "upgrades_file_string": "capacitor", 
                                       "model": CapacitorControllerResultType},
                 "regulator_control": {"identifier": "name", "parameter_list": ["vreg", "band", "delay"], "upgrades_file_string": "regcontrol", 
                                       "model": VoltageRegulatorResultType}, 
    }
    thermal_cost_df = costs_dict["thermal"]
    thermal_cost_df = pd.concat([thermal_cost_df.drop(['equipment_parameters'], axis=1), thermal_cost_df['equipment_parameters'].apply(pd.Series)], axis=1)
    voltage_cost_df = costs_dict["voltage"]
    
    output_file = []
    for stage_item in feeder_stats["stage_results"]:
        if (stage_item["stage"].lower() == "initial") and (stage_item["upgrade_type"].lower() == "thermal"):
            all_original_equipment = stage_item
        if (stage_item["stage"].lower() == "final") and (stage_item["upgrade_type"].lower() == "voltage"):
            all_latest_equipment = stage_item
    
    thermal_summary_df = create_thermal_output_summary(all_original_equipment, all_latest_equipment, thermal_equipment_type_list,
                                                        props_dict, thermal_cost_df, upgrades_dict, output_cols)
    
    voltage_summary_df = create_voltage_output_summary(all_original_equipment, all_latest_equipment, voltage_equipment_type_list,
                                                        props_dict, voltage_cost_df, upgrades_dict, output_cols)
    combined_df = pd.concat([thermal_summary_df, voltage_summary_df])
    combined_df["name"] = kwargs.get("job_name", None)
    return combined_df


def create_opendss_definition(config_definition_dict, action_type="New", property_list=None):
    """This function creates an opendss element definition for any generic equipment

    Returns
    -------
    str
    """
    command_string = f"{action_type} {config_definition_dict['equipment_type']}.{config_definition_dict['name']}"
    logger.debug(f"New {config_definition_dict['equipment_type']}.{config_definition_dict['name']} being defined")
    # these properties contain data (refer OpenDSS manual for more information on these parameters)
    if property_list is None:
        property_list = list(set(config_definition_dict.keys()) - {"name", "equipment_type"})
    empty_field_values = ["----", "nan", "NaN", "None", None, np.nan]
    for property_name in property_list:
        if isinstance(config_definition_dict[property_name], float):
            if np.isnan(config_definition_dict[property_name]):
                continue
        if config_definition_dict[property_name] in empty_field_values:
            continue
        # if the value is not empty and is not nan, only then add it into the command string
        temp_s = f" {property_name}={config_definition_dict[property_name]}"
        command_string = command_string + temp_s
    return command_string


def ensure_line_config_exists(chosen_option, new_config_type, external_upgrades_technical_catalog): 
    """This function check if a line config exists in the network. 
    If it doesn't exist, it checks the external catalog (if available) and returns a new dss definition string.
    
    Returns
    -------
    str
    """
    existing_config_dict = {"linecode": get_line_code(), "geometry": get_line_geometry()}
    new_config_name = chosen_option[new_config_type].lower()
    # if there are no existing config definitions
    if existing_config_dict[new_config_type].empty:
        command_string = add_new_lineconfig_definition(chosen_option, new_config_type, external_upgrades_technical_catalog)
    else:
        # if linecode or linegeometry is not present in existing network definitions
        if not existing_config_dict[new_config_type]["name"].str.lower().isin([new_config_name]).any():  
            command_string = add_new_lineconfig_definition(chosen_option, new_config_type, external_upgrades_technical_catalog)
        else:
            command_string = None   
    return command_string


def add_new_lineconfig_definition(chosen_option, new_config_type, external_upgrades_technical_catalog):
    # add definition for linecode or linegeometry
    if external_upgrades_technical_catalog is None:
        raise UpgradesExternalCatalogRequired(f"External upgrades technical catalog not available to determine line config type")
    if (new_config_type not in external_upgrades_technical_catalog):
        raise UpgradesExternalCatalogMissingObjectDefinition(
            f"{new_config_type} definitions not found in external catalog."
            f" Please check catalog, and add {new_config_type} definitions in it.")
    external_config_df = pd.DataFrame(external_upgrades_technical_catalog[new_config_type])
    if external_config_df.empty: 
        raise UpgradesExternalCatalogMissingObjectDefinition(
            f"{new_config_type} definitions not found in external catalog." 
            f" Please check catalog, and add {new_config_type} definitions in it.")
    new_config_name = chosen_option[new_config_type]
    if external_config_df["name"].str.lower().isin([new_config_name.lower()]).any():
        config_definition_df = external_config_df.loc[external_config_df["name"].str.lower() == new_config_name.lower()].copy()
        if len(config_definition_df) == 1:  # if there is only one definition of that config name
            config_definition_dict = dict(config_definition_df.iloc[0])  
        elif len(config_definition_df) > 1:   # if there is more than one definition of that config name
            config_definition_df["temp_deviation"] = abs(config_definition_df["normamps"] - chosen_option["normamps"])
            config_definition_dict = dict(config_definition_df.loc[config_definition_df["temp_deviation"].idxmin()])
            config_definition_dict.pop("temp_deviation")
        else:  # if definition not found
            raise UpgradesExternalCatalogMissingObjectDefinition(
                f"{new_config_name} definition of {new_config_type} type {new_config_type} not found in external catalog.")
        if config_definition_dict["normamps"] != chosen_option["normamps"]:
            logger.warning(f"Mismatch between noramps for linecode {new_config_name} ({config_definition_dict['normamps']}A) "
                            f"and chosen upgrade option normamps ({chosen_option['normamps']}A): {chosen_option['name']}")
        config_definition_dict["name"] = new_config_name  # to keep same case of config name (for consistency)
        # check format of certain fields, and prepare data to write opendss definition
        matrix_fields = [s for s in config_definition_dict.keys() if 'matrix' in s]
        for field in matrix_fields:
            config_definition_dict[field] = str(config_definition_dict[field]).replace("'","")
            config_definition_dict[field] = config_definition_dict[field].replace("[","(")
            config_definition_dict[field] = config_definition_dict[field].replace("]",")")
        config_definition_dict["equipment_type"] = new_config_type
        command_string = create_opendss_definition(config_definition_dict=config_definition_dict)
    else:
        raise UpgradesExternalCatalogMissingObjectDefinition(
            f"{new_config_type} definition for {new_config_name} not found in external catalog."
        )
    return command_string


def get_present_loading_condition():
    """ Get present loading condition for all loads
    
    Returns
    -------
    DataFrame
    """
    load_dict = {}
    dss.Circuit.SetActiveClass("Load")
    flag = dss.ActiveClass.First()

    while flag > 0:
        # Get the name of the load
        load_dict[dss.CktElement.Name()] = {
                                            'Num_phases': float(dss.Properties.Value("phases")),
                                            'kV': float(dss.Properties.Value("kV")),
                                            'kVA': float(dss.Properties.Value("kVA")),
                                            'kW': float(dss.Properties.Value("kW")),
                                            'pf': dss.Properties.Value("pf"),
                                            'Bus1': dss.Properties.Value("bus1"),
                                            'Powers': dss.CktElement.Powers(),
                                            'NetPower': sum(dss.CktElement.Powers()[::2]),
                                            }
        # Move on to the next Load...
        flag = dss.ActiveClass.Next()
    load_df = pd.DataFrame.from_dict(load_dict, "index")
    return load_df


def get_present_storage_condition():
    """ Get present operating condition for all storage
    
    Returns
    -------
    DataFrame
    """
    storage_dict = {}
    dss.Circuit.SetActiveClass('Storage')
    flag = dss.ActiveClass.First()
    while flag > 0:
        # Get the name of the load
        storage_dict[dss.CktElement.Name()] = {
            'Num_phases': float(dss.Properties.Value("phases")),
            'kV': float(dss.Properties.Value("kV")),
            'kVA': float(dss.Properties.Value("kVA")),
            'kW': float(dss.Properties.Value("kW")),
            'pf': dss.Properties.Value("pf"),
            'Bus1': dss.Properties.Value("bus1"),
            'Powers': dss.CktElement.Powers(),
            'NetPower': sum(dss.CktElement.Powers()[::2]),
        }
        # Move on to the next ...
        flag = dss.ActiveClass.Next()
    storage_df = pd.DataFrame.from_dict(storage_dict, "index")
    return storage_df


def get_present_pvgeneration():
    """ Get present generation for all pv systems
    
    Returns
    -------
    DataFrame
    """
    pv_dict = {}
    dss.Circuit.SetActiveClass("PVSystem")
    flag = dss.ActiveClass.First()
    while flag:
        pv_dict[dss.CktElement.Name()] = {
                                            'Num_phases': float(dss.Properties.Value("phases")),
                                            'kV': float(dss.Properties.Value("kV")),
                                            'kVA': float(dss.Properties.Value("kVA")),
                                            'kvar': float(dss.Properties.Value("kvar")),
                                            'Irradiance': float(dss.Properties.Value("Irradiance")),
                                            'connection': dss.Properties.Value("conn"),
                                            'Pmpp': float(dss.Properties.Value("Pmpp")),
                                            'Powers': dss.CktElement.Powers(),
                                            'NetPower': sum(dss.CktElement.Powers()[::2]),
                                            'pf': dss.Properties.Value("pf"),
                                            'Bus1': dss.Properties.Value("bus1"),
                                            'Voltages': dss.CktElement.Voltages(),
                                            'VoltagesMagAng': dss.CktElement.VoltagesMagAng(),
                                            'VoltagesMag': float(dss.CktElement.VoltagesMagAng()[0]),
                                            }
        flag = dss.ActiveClass.Next() > 0
    pv_df = pd.DataFrame.from_dict(pv_dict, "index")
    return pv_df
   

def check_enabled_property(all_df, element_name):
    """This function checks values for the "enabled" property of a dss object 
    """
    flag = any(ele.lower() in ["yes", "no"] for ele in all_df.enabled.unique())
    if not flag:
        raise OpenDssCompileError(f"Unexpected values {all_df.enabled.unique()} received for 'enabled' {element_name} "
                                  "property. Check OpenDSS version")
    return


def check_switch_property(all_df):
    """This function checks values for the "switch" property of line object 
    """
    flag = any(ele.lower() in ["yes", "no"] for ele in all_df.Switch.unique())
    if not flag:
        raise OpenDssCompileError(f"Unexpected values {all_df.Switch.unique()} received for 'Switch' line "
                                  "property. Check OpenDSS version")
    return


def get_all_transformer_info_instance(upper_limit=None, compute_loading=True):
    """This collects transformer information

    Returns
    -------
    DataFrame
    """
    all_df = dss.utils.class_to_dataframe("transformer")
    check_enabled_property(all_df, element_name="transformer")
    if len(all_df) == 0:
        return pd.DataFrame()
    all_df["name"] = all_df.index.str.split(".").str[1]
    all_df["equipment_type"] = all_df.index.str.split(".").str[0]
    # extract only enabled lines
    all_df = all_df.loc[all_df["enabled"].str.lower() == "yes"]
    all_df["conn"] = all_df["conn"].str.strip()  # remove trailing space from conn field
    # define empty new columns
    all_df['bus_names_only'] = None
    all_df["amp_limit_per_phase"] = np.nan
    all_df[DSS_XFMR_INT_FIELDS] = all_df[DSS_XFMR_INT_FIELDS].astype(int)
    all_df[DSS_XFMR_FLOAT_FIELDS] = all_df[DSS_XFMR_FLOAT_FIELDS].astype(float)
    if compute_loading:
        all_df["max_amp_loading"] = np.nan
        all_df["max_per_unit_loading"] = np.nan
        all_df["status"] = ""
    for index, row in all_df.iterrows():
        all_df.at[index, "kVs"] = [float(a) for a in row["kVs"]]
        all_df.at[index, "kVAs"] = [float(a) for a in row["kVAs"]]
        try:
            all_df.at[index, "Xscarray"] = [float(a) for a in row["Xscarray"]]  # before opendssdirect version 0.7.0
        except ValueError:
            all_df.at[index, "Xscarray"] = [float(a) for a in row["Xscarray"][0].split(" ")]  # in opendssdirect version 0.7.0
        all_df.at[index, "%Rs"] = [float(a) for a in row["%Rs"]]
        all_df.at[index, "taps"] = [float(a) for a in row["taps"]]
        all_df.at[index, "bus_names_only"] = [a.split(".")[0].lower() for a in row["buses"]]
        # first winding is considered primary winding
        primary_kv = float(row["kVs"][0])
        primary_kva = float(row["kVAs"][0])
        if row["phases"] > 1:
            amp_limit_per_phase = primary_kva / (primary_kv * math.sqrt(3))
        elif row["phases"] == 1:
            amp_limit_per_phase = primary_kva / primary_kv
        else:
            raise InvalidOpenDssElementError(f"Incorrect number of phases for transformer {row['name']}")
        all_df.at[index, "amp_limit_per_phase"] = amp_limit_per_phase
        if compute_loading:
            if upper_limit is None:
                raise Exception("Transformer upper limit is to be passed to function to compute transformer loading")
            dss.Circuit.SetActiveElement("Transformer.{}".format(row["name"]))
            extract_magang = dss.CktElement.CurrentsMagAng()[: 2 * row["phases"]]  # extract elements based on num of ph
            xfmr_current_magnitude = extract_magang[::2]
            max_amp_loading = max(xfmr_current_magnitude)
            max_per_unit_loading = round(max_amp_loading / amp_limit_per_phase, 4)
            all_df.at[index, "max_amp_loading"] = max_amp_loading
            all_df.at[index, "max_per_unit_loading"] = max_per_unit_loading
            if max_per_unit_loading > upper_limit:
                all_df.at[index, "status"] = "overloaded"
            elif max_per_unit_loading == 0:
                all_df.at[index, "status"] = "unloaded"
            else:
                all_df.at[index, "status"] = "normal"
    all_df = all_df.reset_index(drop=True).set_index('name')
    return all_df.reset_index()


def add_info_line_definition_type(all_df):
    all_df["line_definition_type"] = "line_definition"
    all_df.loc[all_df["linecode"] != "", "line_definition_type"] = "linecode"
    all_df.loc[all_df["geometry"] != "", "line_definition_type"] = "geometry"
    return all_df


def determine_line_placement(line_series):
    """ Distinguish between overhead and underground cables.
    Latest opendss version has property "LineType"
    line_placement is determined via:
    1. "LineType" property
    2. height property if defined as line geometry
    # line_placement determined via height takes precedence over linetype property
    # this is because if linetype property is not defined in opendss definition, then default: oh is assigned
    
    If line_placement is still not available, it is determined using presence of string "oh" or "ug" in name

    Parameters
    ----------
    line_series

    Returns
    -------
    dict
    """
    info_dict = {}
    info_dict["line_placement"] = None
    line_placement = None
    # use linetype property to determine line_placement
    if ("LineType" in line_series) and (line_series["LineType"] in ["oh", "ug"]):
        if line_series["LineType"] == "oh":
            linetype_placement = "overhead"
        else:
            linetype_placement = "underground"
    line_placement = linetype_placement
    if line_series["line_definition_type"] == "geometry":
        # use height property to determine line_placement
        dss.Circuit.SetActiveClass("linegeometry")
        dss.ActiveClass.Name(line_series["geometry"])
        h = float(dss.Properties.Value("h"))
        info_dict["h"] = 0
        if h >= 0:
            geom_placement = "overhead"
        else:
            geom_placement = "underground"
        # line_placement determined via height takes precedence over linetype property
        # this is because if linetype property is not defined in opendss definition, then default: oh is assigned
        if linetype_placement != geom_placement:
            line_placement = geom_placement 
    # if line_placement is still None, then use line name to determine line placement
    if line_placement is None:
        if ("oh" in line_series["geometry"].lower()) or ("oh" in line_series["linecode"].lower()) :
            line_placement = "overhead"
        elif ("ug" in line_series["geometry"].lower()) or ("ug" in line_series["linecode"].lower()):
            line_placement = "underground"
        else:
            line_placement = "overhead"  # default is taken as overhead
    info_dict["line_placement"] = line_placement
    return info_dict


def get_all_line_info_instance(upper_limit=None, compute_loading=True, ignore_switch=True):
    """This collects line information.
    
    dss.Lines.Units() gives an integer. It can be mapped as below:
    units_config = ["none", "mi", "kft", "km", "m", "Ft", "in", "cm"]  # Units key for lines taken from OpenDSS
    units_config[dss.Lines.Units() - 1]

    Returns
    -------
    DataFrame
    """
    all_df = dss.utils.class_to_dataframe("line")
    if len(all_df) == 0:
        return pd.DataFrame()
    check_enabled_property(all_df, element_name="line")
    all_df["name"] = all_df.index.str.split(".").str[1]
    all_df["equipment_type"] = all_df.index.str.split(".").str[0]
    # extract only enabled lines
    all_df = all_df.loc[all_df["enabled"].str.lower() == "yes"]
    all_df = add_info_line_definition_type(all_df)
    # define empty new columns
    all_df["kV"] = np.nan
    all_df["h"] = np.nan
    all_df["line_placement"] = ""
    all_df[DSS_LINE_INT_FIELDS] = all_df[DSS_LINE_INT_FIELDS].astype(int)
    all_df[DSS_LINE_FLOAT_FIELDS] = all_df[DSS_LINE_FLOAT_FIELDS].astype(float)
    if compute_loading:
        all_df["max_amp_loading"] = np.nan
        all_df["max_per_unit_loading"] = np.nan
        all_df["status"] = ""
    for index, row in all_df.iterrows():
        dss.Circuit.SetActiveBus(row["bus1"])
        kv_b1 = dss.Bus.kVBase()
        dss.Circuit.SetActiveBus(row["bus2"])
        kv_b2 = dss.Bus.kVBase()
        dss.Circuit.SetActiveElement("Line.{}".format(row["name"]))
        if round(kv_b1) != round(kv_b2):
            raise InvalidOpenDssElementError("To and from bus voltages ({} {}) do not match for line {}".format(
                kv_b2, kv_b1, row['name']))
        all_df.at[index, "kV"] = kv_b1
        # Distinguish between overhead and underground cables
        # currently there is no way to distinguish directy using opendssdirect/pydss etc.
        # It is done here using property 'height' parameter and if string present in name
        placement_dict = determine_line_placement(row)
        for key in placement_dict.keys():
            all_df.at[index, key] = placement_dict[key] 
        if row["units"] == "none":
            # possible unit values: {none | mi|kft|km|m|Ft|in|cm } Default is None - assumes length units match impedance units.
            # if units match, then it returns none: in this case, assign value from other lines present in dataframe
            if dss.Lines.Units() != 0:
                all_df.at[index, "units"] = DSS_UNIT_CONFIG[dss.Lines.Units()]
            else:
                def_unit = all_df.units.unique()[0]
                if def_unit != "none":
                    all_df.at[index, "units"] = def_unit
                else:
                    all_df.at[index, "units"] = "m"  # if a unit was not found, assign default of m
        # if line loading is to be computed
        if compute_loading:
            if upper_limit is None:
                raise Exception("Line upper limit is to be passed to function to compute line loading")
            dss.Circuit.SetActiveElement("Line.{}".format(row["name"]))
            extract_magang = dss.CktElement.CurrentsMagAng()[: 2 * row["phases"]]
            line_current = extract_magang[::2]
            max_amp_loading = max(line_current)
            max_per_unit_loading = round(max_amp_loading / row["normamps"], 4)
            all_df.at[index, "max_amp_loading"] = max_amp_loading
            all_df.at[index, "max_per_unit_loading"] = max_per_unit_loading
            if max_per_unit_loading > upper_limit:
                all_df.at[index, "status"] = "overloaded"
            elif max_per_unit_loading == 0:
                all_df.at[index, "status"] = "unloaded"
            else:
                all_df.at[index, "status"] = "normal"
    all_df = all_df.reset_index(drop=True).set_index('name')
    all_df["kV"] = all_df["kV"].round(5)
    # add units to switch length (needed to plot graph). By default, length of switch is taken as max
    check_switch_property(all_df)
    all_df.loc[(all_df.units == 'none') & (all_df.Switch.str.lower() == "yes"), 'units'] = 'm'
    # if switch is to be ignored
    if ignore_switch:
        all_df = all_df.loc[all_df['Switch'].str.lower() == "no"]
    return all_df.reset_index()


def compare_multiple_dataframes(comparison_dict, deciding_column_name, comparison_type="max"):
    """This function compares all dataframes in a given dictionary based on a deciding column name

    Returns
    -------
    Dataframe
    """
    summary_df = pd.DataFrame()
    for df_name in comparison_dict.keys():
        summary_df[df_name] = comparison_dict[df_name][deciding_column_name]
    if comparison_type == "max":
        label_df = summary_df.idxmax(axis=1)  # find dataframe name that has max 
    elif comparison_type == "min":
        label_df = summary_df.idxmax(axis=1)  # find dataframe name that has min 
    else:
        raise Exception(f"Unknown comparison type {comparison_type} passed.")
    final_list = []
    for index, label in label_df.items():  # index is element name
        temp_dict = dict(comparison_dict[label].loc[index])
        temp_dict.update({"name": index})
        final_list.append(temp_dict)
    final_df = pd.DataFrame(final_list)
    return final_df
        

@track_timing(timer_stats_collector)
def get_thermal_equipment_info(compute_loading, equipment_type, upper_limit=None, ignore_switch=False, **kwargs):
    """This function determines the thermal equipment loading (line, transformer), based on timepoint multiplier

    Returns
    -------
    DataFrame
    """
    timepoint_multipliers = kwargs.get("timepoint_multipliers", None)
    multiplier_type = kwargs.get("multiplier_type", LoadMultiplierType.ORIGINAL)
     # if there are no multipliers, run on rated load i.e. multiplier=1. 0
     # if compute_loading is false, then just run once (no need to check multipliers)
    if (timepoint_multipliers is None) or (not compute_loading) or (multiplier_type == LoadMultiplierType.ORIGINAL): 
        if compute_loading and multiplier_type != LoadMultiplierType.ORIGINAL:
            apply_uniform_timepoint_multipliers(multiplier_name=1, field="with_pv", **kwargs)
        if equipment_type == "line":
            loading_df = get_all_line_info_instance(compute_loading=compute_loading, upper_limit=upper_limit, ignore_switch=ignore_switch)
        elif equipment_type == "transformer":
            loading_df = get_all_transformer_info_instance(compute_loading=compute_loading, upper_limit=upper_limit)
        return loading_df
    if multiplier_type == LoadMultiplierType.UNIFORM:
        comparison_dict = {}
        for pv_field in timepoint_multipliers["load_multipliers"].keys():
            logger.debug(pv_field)
            for multiplier_name in timepoint_multipliers["load_multipliers"][pv_field]:
                logger.debug("Multipler name: %s", multiplier_name)
                # this changes the dss network load and pv
                apply_uniform_timepoint_multipliers(multiplier_name=multiplier_name, field=pv_field, **kwargs)
                if equipment_type.lower() == "line":
                    deciding_column_name = "max_per_unit_loading"
                    loading_df = get_all_line_info_instance(compute_loading=compute_loading, upper_limit=upper_limit, ignore_switch=ignore_switch)
                elif equipment_type.lower() == "transformer":
                    deciding_column_name = "max_per_unit_loading"
                    loading_df = get_all_transformer_info_instance(compute_loading=compute_loading, upper_limit=upper_limit)
                loading_df.set_index("name", inplace=True)
                comparison_dict[pv_field+"_"+str(multiplier_name)] = loading_df
        # compare all dataframe, and create one that contains all worst loading conditions (across all multiplier conditions)
        loading_df = compare_multiple_dataframes(comparison_dict, deciding_column_name, comparison_type="max")
    else:
        raise Exception(f"Undefined multiplier_type {multiplier_type} passed.")   
    return loading_df
    

def get_regcontrol_info(correct_PT_ratio=False, nominal_voltage=None):
    """This collects enabled regulator control information.
    If correcting PT ratio, the following information is followed (based on OpenDSS documentation)
    PT ratio: # If the winding is Wye, the line-to-neutral voltage is used. Else, the line-to-line voltage is used.
              # Here, bus kV is taken from Bus.kVBase
    
    Bus base kV:  Returns L-L voltages for 2- and 3-phase. Else for 1-ph, return L-N voltage

    Returns
    -------
    DataFrame
    """
    all_df = dss.utils.class_to_dataframe("regcontrol")
    if len(all_df) == 0:
        return pd.DataFrame()
    check_enabled_property(all_df, element_name="regcontrol")    
    all_df["name"] = all_df.index.str.split(".").str[1]
    all_df["equipment_type"] = all_df.index.str.split(".").str[0]
    float_columns = ['winding', 'vreg', 'band', 'ptratio', 'delay']
    all_df[float_columns] = all_df[float_columns].astype(float)
    all_df['at_substation_xfmr_flag'] = False  # by default, reg control is considered to be not at substation xfmr
    ckt_info_dict = get_circuit_info()
    sub_xfmr_present = False
    sub_xfmr_name = None
    if ckt_info_dict['substation_xfmr'] is not None:
        sub_xfmr_present = True
        sub_xfmr_name = ckt_info_dict['substation_xfmr']['name']
    if correct_PT_ratio:
        if nominal_voltage is None:
            raise Exception("Nominal voltage not provided to correct regcontrol PT ratio.")
        all_df['old_ptratio'] = all_df['ptratio']
        
    for index, row in all_df.iterrows():
        dss.Circuit.SetActiveElement("Regcontrol.{}".format(row["name"]))
        reg_bus = dss.CktElement.BusNames()[0].split(".")[0]
        all_df.at[index, "reg_bus"] = reg_bus
        dss.Circuit.SetActiveBus(reg_bus)
        all_df.at[index, "bus_num_phases"] = dss.CktElement.NumPhases()
        all_df.at[index, "bus_kv"] = dss.Bus.kVBase()
        dss.Circuit.SetActiveElement("Transformer.{}".format(row["transformer"]))
        all_df.at[index, "transformer_kva"] = float(dss.Properties.Value("kva"))
        dss.Transformers.Wdg(1)  # setting winding to 1, to get kV for winding 1
        all_df.at[index, "transformer_kv"] = dss.Transformers.kV()
        all_df.at[index, "transformer_conn"] = dss.Properties.Value("conn").strip()  # opendss returns conn with a space 
        all_df.at[index, "transformer_bus1"] = dss.CktElement.BusNames()[0].split(".")[0]
        all_df.at[index, "transformer_bus2"] = dss.CktElement.BusNames()[1].split(".")[0]
        if correct_PT_ratio:
            if (all_df.loc[index]["bus_num_phases"] > 1) and (all_df.loc[index]["transformer_conn"].lower() == "wye"):
                kV_to_be_used = all_df.loc[index]["transformer_kv"] * 1000 / math.sqrt(3)
            else:
                kV_to_be_used = all_df.loc[index]["transformer_kv"] * 1000
            # kV_to_be_used = dss.Bus.kVBase() * 1000
            all_df.at[index, "ptratio"] = kV_to_be_used / nominal_voltage
        if sub_xfmr_present and (row["transformer"] == sub_xfmr_name):  # if reg control is at substation xfmr
            all_df.at[index, 'at_substation_xfmr_flag'] = True
    all_df = all_df.reset_index(drop=True).set_index('name')        
    all_df = all_df.loc[all_df['enabled'].str.lower() == "yes"]
    return all_df.reset_index()


def get_capacitor_info(nominal_voltage=None, correct_PT_ratio=False):
    """
    This collects capacitor information.
    For correcting PT ratio, the following information and definitions are followed:
    # cap banks are 3 phase, 2 phase or 1 phase. 1 phase caps will have LN voltage
    # PT ratio: Ratio of the PT that converts the monitored voltage to the control voltage. 
    # If the capacitor is Wye, the 1st phase line-to-neutral voltage is monitored.
    # Else, the line-to-line voltage (1st - 2nd phase) is monitored.
    # Capacitor kv: Rated kV of the capacitor (not necessarily same as bus rating). 
    # For Phases=2 or Phases=3, it is line-to-line (phase-to-phase) rated voltage. 
    # For all other numbers of phases, it is actual rating. (For Delta connection this is always line-to-line rated voltage). 
    This function doesnt currently check if object is "enabled".
    
    Returns
    -------
    DataFrame
    """
    all_df = dss.utils.class_to_dataframe("capacitor")
    if len(all_df) == 0:
        return pd.DataFrame()
    check_enabled_property(all_df, element_name="capacitor")  
    all_df["capacitor_name"] = all_df.index.str.split(".").str[1]
    all_df["equipment_type"] = all_df.index.str.split(".").str[0]
    float_columns = ["phases", "kv"]
    all_df[float_columns] = all_df[float_columns].astype(float)
    all_df = all_df.reset_index(drop=True).set_index("capacitor_name")
    # collect capcontrol information to combine with capcontrols
    capcontrol_df = get_cap_control_info()
    capcontrol_df.rename(columns={'name': 'capcontrol_name', 'capacitor': 'capacitor_name', 'type': 'capcontrol_type',
                                  'equipment_type': 'capcontrol_present'}, inplace=True)
    capcontrol_df = capcontrol_df.set_index("capacitor_name")
    # with capacitor name as index, concatenate capacitor information with cap controls
    all_df = pd.concat([all_df, capcontrol_df], axis=1)
    all_df.index.name = 'capacitor_name'
    all_df = all_df.reset_index().set_index('capacitor_name')
    
    if correct_PT_ratio and (len(capcontrol_df) > 0):
        if nominal_voltage is None:
            raise Exception("Nominal voltage not provided to correct capacitor bank PT ratio.")
        all_df['old_PTratio'] = all_df['PTratio']
    
    # iterate over all capacitors
    for index, row in all_df.iterrows():
        all_df.at[index, "kvar"] = [float(a) for a in row["kvar"]][0]
        # if capcontrol type is empty, then that capacitor does not have controls
        # correct PT ratios for existing cap controls
        if correct_PT_ratio and (len(capcontrol_df) > 0):
            if row["phases"] > 1 and row["conn"].lower() == "wye":
                kv_to_be_used = (row['kv'] * 1000) / math.sqrt(3)
            else:
                kv_to_be_used = row['kv'] * 1000
            all_df.at[index, "PTratio"] = kv_to_be_used / nominal_voltage
    return all_df.reset_index()


def get_cap_control_info():
    """This collects capacitor control information

    Returns
    -------
    DataFrame
    """
    all_df = dss.utils.class_to_dataframe("capcontrol")
    if len(all_df) == 0:
        capcontrol_columns = ["name",  "equipment_type", "element", "terminal", "capacitor", 
                              "type", "PTratio", "CTratio", "ONsetting", "OFFsetting", "Delay", 
                              "VoltOverride", "Vmax", "Vmin", "DelayOFF", "DeadTime", "CTPhase", 
                              "PTPhase", "VBus", "EventLog", "UserModel", "UserData", "pctMinkvar", 
                              "Reset", "basefreq", "enabled", "like"]
        return pd.DataFrame(columns=capcontrol_columns)
    check_enabled_property(all_df, element_name="capcontrol")
    all_df["name"] = all_df.index.str.split(".").str[1]
    all_df["equipment_type"] = all_df.index.str.split(".").str[0]
    CAPCONTROL_FLOAT_FIELDS = ["CTratio", "DeadTime", "Delay", "DelayOFF", "OFFsetting", "ONsetting", "PTratio",
                              "Vmax", "Vmin"]
    all_df[CAPCONTROL_FLOAT_FIELDS] = all_df[CAPCONTROL_FLOAT_FIELDS].astype(float)
    all_df = all_df.reset_index(drop=True).set_index("name")
    return all_df.reset_index()


def get_line_geometry():
    """This collects all line geometry information

    Returns
    -------
    DataFrame
    """
    active_class_name = 'linegeometry'
    all_df = dss.utils.class_to_dataframe(active_class_name)
    if len(all_df) == 0:
        return pd.DataFrame()
    all_df['name'] = all_df.index.str.split('.').str[1]
    all_df['equipment_type'] = all_df.index.str.split('.').str[0]
    all_df.reset_index(inplace=True, drop=True)
    all_df[DSS_LINEGEOMETRY_FLOAT_FIELDS] = all_df[DSS_LINEGEOMETRY_FLOAT_FIELDS].astype("float")
    all_df[DSS_LINEGEOMETRY_INT_FIELDS] = all_df[DSS_LINEGEOMETRY_INT_FIELDS].astype("int")
    all_df = all_df[list(LineGeometryCatalogModel.schema(True).get("properties").keys())]
    return all_df


def get_line_code():
    """This collects all line codes information

    Returns
    -------
    DataFrame
    """
    active_class_name = 'linecode'
    all_df = dss.utils.class_to_dataframe(active_class_name)
    if len(all_df) == 0:
        return pd.DataFrame()
    all_df['name'] = all_df.index.str.split('.').str[1]
    all_df['equipment_type'] = all_df.index.str.split('.').str[0]
    all_df.reset_index(inplace=True, drop=True)
    all_df[DSS_LINECODE_FLOAT_FIELDS] = all_df[DSS_LINECODE_FLOAT_FIELDS].astype("float")
    all_df[DSS_LINECODE_INT_FIELDS] = all_df[DSS_LINECODE_INT_FIELDS].astype("int")
    all_df = all_df[list(LineCodeCatalogModel.schema(True).get("properties").keys())]
    return all_df


def get_wire_data():
    """This collects all wire data information

    Returns
    -------
    DataFrame
    """
    active_class_name = 'wiredata'
    all_df = dss.utils.class_to_dataframe(active_class_name)
    if len(all_df) == 0:
        return pd.DataFrame()
    all_df['name'] = all_df.index.str.split('.').str[1]
    all_df['equipment_type'] = all_df.index.str.split('.').str[0]
    all_df.reset_index(inplace=True, drop=True)
    return all_df


def get_cn_data():
    """This collects all cn data information

    Returns
    -------
    DataFrame
    """
    active_class_name = 'cndata'
    all_df = dss.utils.class_to_dataframe(active_class_name)
    if len(all_df) == 0:
        return pd.DataFrame()
    all_df['name'] = all_df.index.str.split('.').str[1]
    all_df['equipment_type'] = all_df.index.str.split('.').str[0]
    all_df.reset_index(inplace=True, drop=True)
    return all_df


def check_dss_run_command(command_string):
    """Runs dss command
    And checks for exception

    Parameters
    ----------
    command_string : str
        dss command to be run

    Raises
    -------
    OpenDssCompileError
        Raised if the command fails

    """
    logger.debug(f"Running DSS command: {command_string}")
    result = dss.Text.Command(f"{command_string}")
    if result is not None:
        raise OpenDssCompileError(f"OpenDSS run_command failed with message: {result}. \nCommand: {command_string}")


@track_timing(timer_stats_collector)
def get_bus_voltages(voltage_upper_limit, voltage_lower_limit, raise_exception=True, **kwargs):
    """This function determines the voltages, based on timepoint multiplier

    Returns
    -------
    DataFrame
    """
    timepoint_multipliers = kwargs.get("timepoint_multipliers", None)
    multiplier_type = kwargs.get("multiplier_type", LoadMultiplierType.ORIGINAL)
     # if there are no multipliers, run on rated load i.e. multiplier=1. 0
     # if compute_loading is false, then just run once (no need to check multipliers)
    if (timepoint_multipliers is None) or (multiplier_type == LoadMultiplierType.ORIGINAL): 
        if multiplier_type != LoadMultiplierType.ORIGINAL:
            apply_uniform_timepoint_multipliers(multiplier_name=1, field="with_pv", **kwargs)
            # determine voltage violations after changes
        bus_voltages_df, undervoltage_bus_list, overvoltage_bus_list, buses_with_violations = get_bus_voltages_instance(
            voltage_upper_limit=voltage_upper_limit, voltage_lower_limit=voltage_lower_limit, raise_exception=raise_exception, 
            **kwargs)
        return bus_voltages_df, undervoltage_bus_list, overvoltage_bus_list, buses_with_violations
    if multiplier_type == LoadMultiplierType.UNIFORM:
        comparison_dict = {}
        for pv_field in timepoint_multipliers["load_multipliers"].keys():
            logger.debug(pv_field)
            for multiplier_name in timepoint_multipliers["load_multipliers"][pv_field]:
                logger.debug("Multipler name: %s", multiplier_name)
                # this changes the dss network load and pv
                apply_uniform_timepoint_multipliers(multiplier_name=multiplier_name, field=pv_field, **kwargs)
                bus_voltages_df, undervoltage_bus_list, overvoltage_bus_list, buses_with_violations = get_bus_voltages_instance(
                    voltage_upper_limit=voltage_upper_limit, voltage_lower_limit=voltage_lower_limit, raise_exception=raise_exception, **kwargs)
                bus_voltages_df.set_index("name", inplace=True)
                comparison_dict[pv_field+"_"+str(multiplier_name)] = bus_voltages_df
        # compare all dataframe, and create one that contains all worst loading conditions (across all multiplier conditions)
        deciding_column_dict = {"max_per_unit_voltage": "max", "min_per_unit_voltage": "min"}
        bus_voltages_df, undervoltage_bus_list, overvoltage_bus_list, buses_with_violations = compare_multiple_dataframes_voltage(comparison_dict=comparison_dict, 
                                                                                                                                  deciding_column_dict=deciding_column_dict,
                                                                                                                                  voltage_upper_limit=voltage_upper_limit,
                                                                                                                                  voltage_lower_limit=voltage_lower_limit)
    else:
        raise Exception(f"Undefined multiplier_type {multiplier_type} passed.")    
    return bus_voltages_df, undervoltage_bus_list, overvoltage_bus_list, buses_with_violations
    

@track_timing(timer_stats_collector)
def get_bus_voltages_instance(voltage_upper_limit, voltage_lower_limit, raise_exception=True, **kwargs):
    """This computes per unit voltages for all buses in network

    Returns
    -------
    DataFrame
    """
    circuit_solve_and_check(raise_exception=raise_exception, **kwargs)  # this is added as a final check for convergence
    all_dict = {}
    all_bus_names = dss.Circuit.AllBusNames()
    for bus_name in all_bus_names:
        dss.Circuit.SetActiveBus(bus_name)
        data_dict = {
            "name": bus_name,
            "voltages": dss.Bus.puVmagAngle()[::2],
            # "kvbase": dss.Bus.kVBase(),
        }
        data_dict["max_per_unit_voltage"] = max(data_dict["voltages"])
        data_dict["min_per_unit_voltage"] = min(data_dict["voltages"])
        data_dict['phase_imbalance'] = data_dict["max_per_unit_voltage"] - data_dict["min_per_unit_voltage"]

        # check for overvoltage violation
        if data_dict["max_per_unit_voltage"] > voltage_upper_limit:
            data_dict['overvoltage_violation'] = True
            data_dict["max_voltage_deviation"] = data_dict["max_per_unit_voltage"] - voltage_upper_limit
        else:
            data_dict['overvoltage_violation'] = False
            data_dict["max_voltage_deviation"] = 0.0

        # check for undervoltage violation
        if data_dict["min_per_unit_voltage"] < voltage_lower_limit:
            data_dict['undervoltage_violation'] = True
            data_dict["min_voltage_deviation"] = voltage_lower_limit - data_dict["min_per_unit_voltage"]
        else:
            data_dict['undervoltage_violation'] = False
            data_dict["min_voltage_deviation"] = 0.0
        all_dict[data_dict["name"]] = data_dict

    all_df = pd.DataFrame.from_dict(all_dict, orient='index').reset_index(drop=True)
    undervoltage_bus_list = list(all_df.loc[all_df['undervoltage_violation'] == True]['name'].unique())
    overvoltage_bus_list = list(all_df.loc[all_df['overvoltage_violation'] == True]['name'].unique())
    buses_with_violations = list(set(undervoltage_bus_list + overvoltage_bus_list))
    return all_df, undervoltage_bus_list, overvoltage_bus_list, buses_with_violations


def compare_multiple_dataframes_voltage(comparison_dict, deciding_column_dict, voltage_upper_limit, voltage_lower_limit):
    """This function compares all dataframes in a given dictionary based on a deciding column 

    Returns
    -------
    Dataframe
    """
    all_df = pd.DataFrame()
    for deciding_column_name in deciding_column_dict.keys():
        summary_df = pd.DataFrame()
        comparison_type = deciding_column_dict[deciding_column_name]
        for df_name in comparison_dict.keys():
            label_df = pd.DataFrame()
            summary_df[df_name] = comparison_dict[df_name][deciding_column_name]
            if comparison_type == "max":
                label_df[deciding_column_name] = summary_df.idxmax(axis=1)  # find dataframe name that has max 
            elif comparison_type == "min":
               label_df[deciding_column_name] = summary_df.idxmin(axis=1)  # find dataframe name that has min 
            else:
                raise Exception(f"Unknown comparison type {comparison_type} passed.")
        final_list = []
        for index, row in label_df.iterrows():  # index is element name
            label = row[deciding_column_name]
            temp_dict = {deciding_column_name: comparison_dict[label].loc[index][deciding_column_name]}
            temp_dict.update({"name": index})
            final_list.append(temp_dict)
        temp_df = pd.DataFrame(final_list)
        temp_df.set_index("name", inplace=True)
        all_df = pd.concat([all_df, temp_df], axis=1)
    bus_voltages_df, undervoltage_bus_list, overvoltage_bus_list, buses_with_violations = get_voltage_violations(voltage_upper_limit=voltage_upper_limit, 
                                                                                                voltage_lower_limit=voltage_lower_limit, 
                                                                                                bus_voltages_df=all_df)
    return bus_voltages_df, undervoltage_bus_list, overvoltage_bus_list, buses_with_violations
        
        
def get_voltage_violations(voltage_upper_limit, voltage_lower_limit, bus_voltages_df):
    """Function to determine voltage violations
    """
    bus_voltages_df['overvoltage_violation'] = False
    bus_voltages_df['undervoltage_violation'] = False
    bus_voltages_df['max_voltage_deviation'] = 0.0
    bus_voltages_df['min_voltage_deviation'] = 0.0
    
    for index, row in bus_voltages_df.iterrows():
        # check for overvoltage violation
        if row["max_per_unit_voltage"] > voltage_upper_limit:
            bus_voltages_df.at[index, 'overvoltage_violation'] = True
            bus_voltages_df.at[index, "max_voltage_deviation"] = row["max_per_unit_voltage"] - voltage_upper_limit
        else:
            bus_voltages_df.at[index, 'overvoltage_violation'] = False
            bus_voltages_df.at[index, "max_voltage_deviation"] = 0.0

        # check for undervoltage violation
        if row["min_per_unit_voltage"] < voltage_lower_limit:
            bus_voltages_df.at[index, 'undervoltage_violation'] = True
            bus_voltages_df.at[index, "min_voltage_deviation"] = voltage_lower_limit - row["min_per_unit_voltage"]
        else:
            bus_voltages_df.at[index, 'undervoltage_violation'] = False
            bus_voltages_df.at[index, "min_voltage_deviation"] = 0.0
    
    bus_voltages_df.reset_index(inplace=True)
    undervoltage_bus_list = list(bus_voltages_df.loc[bus_voltages_df['undervoltage_violation'] == True]['name'].unique())
    overvoltage_bus_list = list(bus_voltages_df.loc[bus_voltages_df['overvoltage_violation'] == True]['name'].unique())
    buses_with_violations = list(set(undervoltage_bus_list + overvoltage_bus_list))
    return bus_voltages_df, undervoltage_bus_list, overvoltage_bus_list, buses_with_violations
            

def determine_available_line_upgrades(line_loading_df):
    """This function creates a dataframe of available line upgrades by dropping duplicates from line dataframe passed.
    """
    all_property_list = list(LineCatalogModel.schema(True).get("properties").keys())
    determining_property_list =  _extract_specific_model_properties_(model_name=LineCatalogModel, field_type_key="determine_upgrade_option", field_type_value=True)
    line_loading_df["kV"] = line_loading_df["kV"].round(5)
    if 'line_definition_type' not in line_loading_df.columns:  # add line_definition_type if not present
        line_loading_df = add_info_line_definition_type(line_loading_df)
    if 'line_placement' not in line_loading_df.columns:
        for index, row in line_loading_df.iterrows():  # add line_placement and h if not present
            info_dict = determine_line_placement(row)
            for key in info_dict.keys():
                line_loading_df.at[index, key] = info_dict[key] 
    line_upgrade_options = line_loading_df[all_property_list]
    # remove duplicate line upgrade options (that might have a different name, but same parameters)
    line_upgrade_options = line_upgrade_options.loc[line_upgrade_options.astype(str).drop_duplicates(subset=determining_property_list, keep="first").index]
    line_upgrade_options.reset_index(drop=True, inplace=True)
    if not line_upgrade_options["name"].is_unique:  # if line upgrade option names are not unique, create new names
        line_upgrade_options = line_upgrade_options.reset_index().rename(columns={'index': 'name'})
        line_upgrade_options['name'] = 'line_' + line_upgrade_options['name'].astype(str)
    return line_upgrade_options[all_property_list]


def determine_available_xfmr_upgrades(xfmr_loading_df):
    """This function creates a dataframe of available transformer upgrades by dropping duplicates from transformer dataframe passed.
    Input dataframe will need to contain "amp_limit_per_phase" column. So if external catalog is supplied, ensure it contains that column.
    """
    all_property_list = list(TransformerCatalogModel.schema(True).get("properties").keys())
    determining_property_list =  _extract_specific_model_properties_(model_name=TransformerCatalogModel, field_type_key="determine_upgrade_option", field_type_value=True)
    xfmr_upgrade_options = xfmr_loading_df[all_property_list]
    xfmr_upgrade_options = xfmr_upgrade_options.loc[xfmr_upgrade_options.astype(str).drop_duplicates(subset=determining_property_list, keep="first").index]
    xfmr_upgrade_options.reset_index(drop=True, inplace=True)
    if not xfmr_upgrade_options["name"].is_unique:  # if xfmr upgrade option names are not unique, create new names
        xfmr_upgrade_options = xfmr_upgrade_options.reset_index().rename(columns={'index': 'name'})
        xfmr_upgrade_options['name'] = 'xfmr_' + xfmr_upgrade_options['name'].astype(str)
    return xfmr_upgrade_options


def remove_duplicate_line_upgrades(line_upgrades_df):
    if line_upgrades_df.empty:
        return line_upgrades_df
    upgrades_subset = line_upgrades_df.loc[line_upgrades_df["action"] == "add"]
    duplicate_equip = list(upgrades_subset[upgrades_subset.final_equipment_name.duplicated()].final_equipment_name.unique())
    if not duplicate_equip:
        return line_upgrades_df
    upgrades_subset = upgrades_subset.groupby('final_equipment_name', group_keys=False).apply(lambda df: df.fillna(method='ffill'))
    upgrades_subset = upgrades_subset.loc[upgrades_subset["final_equipment_name"].duplicated(keep=False)]
    duplicates_final = upgrades_subset.loc[upgrades_subset["upgrade_type"] == "upgrade"]
    duplicates_final = duplicates_final.sort_values("normamps", ascending=False)
    duplicates_final = duplicates_final.drop_duplicates(subset=["upgrade_type", "action", "final_equipment_name"], keep="first")  # drop any duplicate rows, keeping highest rating
    duplicates_final["upgrade_type"] = "new_parallel"
    duplicates_final = duplicates_final.drop(columns=["original_equipment_name"])
    # get original equipment name : 
    og = upgrades_subset.loc[upgrades_subset["upgrade_type"] == "new_parallel"].set_index("final_equipment_name")[["original_equipment_name"]]
    duplicates_final = pd.concat([duplicates_final.set_index("final_equipment_name"), og], axis=1).reset_index()
    
    # keep all non-duplicated
    final_line_upgrades_df = line_upgrades_df.loc[~(line_upgrades_df["final_equipment_name"].isin(duplicate_equip))]
    final_line_upgrades_df = pd.concat([final_line_upgrades_df, duplicates_final])
    return final_line_upgrades_df
    
    
def remove_duplicate_transformer_upgrades(xfmr_upgrades_df):
    if xfmr_upgrades_df.empty:
        return xfmr_upgrades_df
    upgrades_subset = xfmr_upgrades_df.loc[xfmr_upgrades_df["action"] == "add"]
    duplicate_equip = list(upgrades_subset[upgrades_subset.final_equipment_name.duplicated()].final_equipment_name.unique())
    if not duplicate_equip:
        return xfmr_upgrades_df
    upgrades_subset = upgrades_subset.groupby('final_equipment_name', group_keys=False).apply(lambda df: df.fillna(method='ffill'))
    upgrades_subset = upgrades_subset.loc[upgrades_subset["final_equipment_name"].duplicated(keep=False)]
    duplicates_final = upgrades_subset.loc[upgrades_subset["upgrade_type"] == "upgrade"]
    duplicates_final = duplicates_final.sort_values("kVA", ascending=False)
    duplicates_final = duplicates_final.drop_duplicates(subset=["upgrade_type", "action", "final_equipment_name"], keep="first")  # drop any duplicate rows, keeping highest rating
    duplicates_final["upgrade_type"] = "new_parallel"
    duplicates_final = duplicates_final.drop(columns="original_equipment_name")
    # get original equipment name : 
    og = upgrades_subset.loc[upgrades_subset["upgrade_type"] == "new_parallel"].set_index("final_equipment_name")[["original_equipment_name"]]
    duplicates_final = pd.concat([duplicates_final.set_index("final_equipment_name"), og], axis=1).reset_index()
    
    # keep all non-duplicated
    final_xfmr_upgrades_df = xfmr_upgrades_df.loc[~(xfmr_upgrades_df["final_equipment_name"].isin(duplicate_equip))]
    final_xfmr_upgrades_df = pd.concat([final_xfmr_upgrades_df, duplicates_final])
    return final_xfmr_upgrades_df


def get_pv_buses(dss):
    pv_buses = []
    flag = dss.PVsystems.First()
    while flag > 0:
        pv_buses.append(dss.Properties.Value('bus1').split('.')[0])
        flag = dss.PVsystems.Next()
    return pv_buses


def get_load_buses(dss):
    load_buses = []
    flag = dss.Loads.First()
    while flag > 0:
        load_buses.append(dss.Properties.Value('bus1').split('.')[0])
        flag = dss.Loads.Next()
    return load_buses


def get_bus_coordinates():
    """This function creates a dataframe of all buses in the circuit with their x and y coordinates

    Returns
    -------

    """
    all_bus_names = dss.Circuit.AllBusNames()
    buses_list = []
    for b in all_bus_names:
        bus_dict = {}
        dss.Circuit.SetActiveBus(b)
        bus_dict['bus_name'] = b.lower()
        bus_dict['x_coordinate'] = dss.Bus.X()
        bus_dict['y_coordinate'] = dss.Bus.Y()
        buses_list.append(bus_dict)
    bus_coordinates_df = pd.DataFrame(buses_list)
    if all(bus_coordinates_df["x_coordinate"].unique() == [0]) and all( bus_coordinates_df["y_coordinate"].unique() == [0]):
        logger.info("Buscoordinates not provided for feeder model.")
    return bus_coordinates_df


def convert_summary_dict_to_df(summary_dict):
    df = pd.DataFrame.from_dict(summary_dict, orient='index')
    df.index.name = "stage"
    return df


def filter_dictionary(dict_data, wanted_keys):
    return {k: dict_data.get(k, None) for k in wanted_keys}


def compare_dict(old, new, properties_to_check=None):
    """function to compare two dictionaries with same format. 
    Only compares common elements present in both original and new dictionaries
    
    """
    field_list = []
    change = {}
    sharedKeys = set(old.keys()).intersection(new.keys())
    if not sharedKeys:  # if there are no shared keys, then exit function
        return change
    all_properties = old[list(sharedKeys)[0]].keys()
    if properties_to_check is None:
        # get all properties from first element of dictionary
        properties_to_check = all_properties
    else:
        properties_to_check = list(set(all_properties) & set(properties_to_check))
    for key in sharedKeys:
        change_flag = False
        for sub_field in properties_to_check:
            if pd.isna(old[key][sub_field]) and pd.isna(new[key][sub_field]):
                continue
            if old[key][sub_field] != new[key][sub_field]:
                change_flag = True
                field_list.append(sub_field)
        if change_flag:
            change[key] = field_list
    return change


def create_timepoint_multipliers_dict(timepoint_multipliers):
    """Creates a dictionary with new load rating, for every property and multiplier.
    Currently, it only does this for loads. But can be modified to accommodate other elements like PV as well.
    In raw_dict, value can be accessed as follows:
    value = raw_dict[property_name][object_name][multiplier_name]
    
    In reformatted_dict (which is returned from this function), value can be accessed as follows:
    value = raw_dict[object_name][property_name][multiplier_name]
    This value will need to be assigned to the object and run.
    This hasnt been used yet.
    
    Returns
    -------
    dict
    """
    for field in timepoint_multipliers.keys():
        if field == "load_multipliers":
            property_list = ["kW"]
            object_name = "Load"
            multiplier_list = []
            # get combined list of multipliers
            for key, value in timepoint_multipliers[field].items():
                multiplier_list = multiplier_list + value
            df = dss.utils.class_to_dataframe(object_name)
            df.reset_index(inplace=True)
            df['name'] = df['index'].str.split(".", expand=True)[1]
            name_list = list(df['name'].values)
            del df["index"]
            df.set_index('name', inplace=True)
            raw_dict = {}
            for property in property_list:
                logger.debug(property)
                df[property] = df[property].astype(float)
                new_df = pd.DataFrame(index=name_list, columns=multiplier_list)
                new_df.index.name = 'name'
                for multiplier in multiplier_list:
                    logger.debug(multiplier)
                    new_df[multiplier] = df[property] * multiplier
                raw_dict[property] = new_df.T.to_dict()
            # reformat dictionary to create desired format
            reformatted_dict = {}
            for name in name_list:
                reformatted_dict[name] = {}
                for property in property_list:
                    reformatted_dict[name][property] = raw_dict[property][name]
        else:
            raise Exception(f"Timepoint multiplier has Unsupported key: {field}. Presently, key 'load_multipliers' is supported.")
    return reformatted_dict


@track_timing(timer_stats_collector)
def apply_timepoint_multipliers_dict(reformatted_dict, multiplier_name, property_list=None, field="load_multipliers",
                                     **kwargs):
    """This uses a dictionary with the format of output received from create_timepoint_multipliers_dict
    Currently, it only does works loads. But can be modified to accommodate other elements like PV as well.

    In input dict: value can be accessed as follows:
    value = raw_dict[object_name][property_name][multiplier_name]
    In this function, value will be assigned to corresponding property and run.
    This hasnt been used yet.
    
    Returns
    -------
    dict
    """
    name_list = list(reformatted_dict.keys())
    if property_list is None:
        property_list = list(reformatted_dict[name_list[0]].keys())
    if field == "load_multipliers":
        flag = dss.Loads.First()
        while flag > 0:
            flag = dss.Loads.Next()
            name = dss.Loads.Name()
            if name not in name_list:  # if load name is not present in dictionary keys, continue
                continue
            for property in property_list:
                value = reformatted_dict[name][property][multiplier_name]
                if property == "kW":
                    dss.Loads.kW(value)
                else:
                    raise Exception(f"Property {property} not defined in multipliers dict")
        circuit_solve_and_check(raise_exception=True, **kwargs)
    else:
        raise Exception(f"Unsupported key in dictionary. Presently, load_multipliers is supported.")
    return reformatted_dict


def apply_uniform_timepoint_multipliers(multiplier_name, field, **kwargs):
    """This function applies a uniform mulitplier to all elements. 
    Currently, the multiplier only does works on loads. But can be modified to accommodate other elements like PV as well.
    It has two options, 1) all pv is enabled. 2) all pv is disabled.
    
    Returns
    -------
    bool
    """
    if field == "with_pv":
        check_dss_run_command("BatchEdit PVSystem..* Enabled=Yes")
    elif field == "without_pv":    
        check_dss_run_command("BatchEdit PVSystem..* Enabled=No")
    else:
        raise Exception(f"Unknown parameter {field} passed in uniform timepoint multiplier dict."
                        f"Acceptable values are 'with_pv', 'without_pv'")
    check_dss_run_command(f"set LoadMult = {multiplier_name}")
    circuit_solve_and_check(raise_exception=True, **kwargs)
    return True



"""CLI commands for transforming source data"""

import logging
import os
import sys

import click

from jade.loggers import setup_logging
from jade.utils.utils import get_cli_string
from disco.sources.factory import list_subcommands, make_source_model


logger = logging.getLogger(__name__)


class TransformCli(click.MultiCommand):
    """Custom CLI to dynamically detect subcommands from an input argument."""

    def __init__(self, *args, **kwargs):
        super(TransformCli, self).__init__(*args, **kwargs)
        self.input_path = None
        self.source_type = None

    def list_commands(self, ctx):
        self.check_input_path()
        return self.source_type.list_transform_subcommands()

    def get_command(self, ctx, name):
        self.check_input_path()
        if name in ("-h", "--help"):
            self.show_subcommand_help_and_exit()

        if name not in self.list_commands(ctx):
            raise click.BadArgumentUsage(f"{name} is not a valid command")
        return self.source_type.get_transform_subcommand(name)

    def check_input_path(self):
        if self.source_type is None:
            show_help_and_exit()

    def show_subcommand_help_and_exit(self):
        commands = self.source_type.list_transform_subcommands()
        input_path = self.input_path
        print("\nAvailable analysis types: {}\n".format(" ".join(commands)))
        print("For additional help run one of the following:")
        for cmd in commands:
            print(f"    disco transform-model {input_path} {cmd} --help")
        sys.exit(0)


def input_path_cb(ctx, param, value):
    ctx.command.input_path = value
    ctx.command.source_type = make_source_model(value)
    return value


def show_help_and_exit():
    print(
        "Transforms source data specified by INPUT_PATH into DISCO models.\n"
        "Subcommands are dependent on the source data type.\n"
    )
    print("Usage: disco transform-model [OPTIONS] INPUT_PATH COMMAND [ARGS]...")
    print("\ninput_path must be specified to determine available commands\n")
    print("Examples:\n")
    print("disco transform-model ~/source-data --help")
    print("disco transform-model ~/source-data snapshot --help")
    print(
        "disco transform-model ~/source-data snapshot --output snapshot-models"
    )
    sys.exit(0)


@click.command(cls=TransformCli)
@click.argument("input_path", callback=input_path_cb)
def transform_model(input_path):
    """Transform input data into a DISCO model"""
    setup_logging("disco", "transform_model.log", mode="a", packages=["disco"])
    logger.info(get_cli_string())



import json
import logging
import os
import time

from jade.utils.timing_utils import track_timing, Timer
from jade.utils.utils import load_data, dump_data

from .fixed_upgrade_parameters import (
    DEFAULT_CAPACITOR_SETTINGS,
    DEFAULT_SUBLTC_SETTINGS,
    DEFAULT_REGCONTROL_SETTINGS
)
from .voltage_upgrade_functions import *
from disco.enums import LoadMultiplierType
from disco.models.upgrade_cost_analysis_generic_output_model import UpgradeViolationResultModel, AllUpgradesTechnicalResultModel
from disco import timer_stats_collector

logger = logging.getLogger(__name__)


@track_timing(timer_stats_collector)
def determine_voltage_upgrades(
    job_name,
    master_path,
    enable_pydss_solve,
    pydss_volt_var_model,
    thermal_config,
    voltage_config,
    thermal_upgrades_dss_filepath,
    voltage_upgrades_dss_filepath,
    upgraded_master_dss_filepath,
    output_json_voltage_upgrades_filepath,
    feeder_stats_json_file,
    voltage_upgrades_directory,
    overall_output_summary_filepath,
    dc_ac_ratio,
    ignore_switch=True,
    verbose=False
):
    start_time = time.time()
    logger.info(f"Simulation Start time: {start_time}")
    timepoint_multipliers = voltage_config["timepoint_multipliers"]
    if timepoint_multipliers is not None:
        multiplier_type = LoadMultiplierType.UNIFORM
    else:
        multiplier_type = LoadMultiplierType.ORIGINAL
    create_plots = voltage_config["create_plots"]
    # default_capacitor settings and customization
    default_capacitor_settings = DEFAULT_CAPACITOR_SETTINGS
    default_capacitor_settings["capON"] = round(
        (
            voltage_config["nominal_voltage"] -
            voltage_config["capacitor_sweep_voltage_gap"] / 2
        ),
        1,
    )
    default_capacitor_settings["capOFF"] = round(
        (
            voltage_config["nominal_voltage"] +
            voltage_config["capacitor_sweep_voltage_gap"] / 2
        ),
        1,
    )
    # default subltc settings and customization
    default_subltc_settings = DEFAULT_SUBLTC_SETTINGS
    default_subltc_settings["vreg"] = 1.03 * voltage_config["nominal_voltage"]
    
    default_regcontrol_settings = DEFAULT_REGCONTROL_SETTINGS
    default_regcontrol_settings["vreg"] = voltage_config["nominal_voltage"]

    if not os.path.exists(thermal_upgrades_dss_filepath):
        raise Exception(f"AutomatedThermalUpgrade did not produce thermal upgrades dss file")
    
    initial_simulation_params = {"enable_pydss_solve": enable_pydss_solve, "pydss_volt_var_model": pydss_volt_var_model,
                                 "dc_ac_ratio": dc_ac_ratio, "max_control_iterations": voltage_config["max_control_iterations"]}
    initial_dss_file_list = [master_path, thermal_upgrades_dss_filepath]
    simulation_params = reload_dss_circuit(dss_file_list=initial_dss_file_list, commands_list=None, **initial_simulation_params)
    simulation_params.update({"timepoint_multipliers": timepoint_multipliers, "multiplier_type": multiplier_type})
    # reading original objects (before upgrades)
    orig_ckt_info = get_circuit_info()
    orig_xfmrs_df =  get_thermal_equipment_info(compute_loading=False, equipment_type="transformer")
    orig_regcontrols_df = get_regcontrol_info(correct_PT_ratio=True, nominal_voltage=voltage_config["nominal_voltage"])
    orig_capacitors_df = get_capacitor_info(correct_PT_ratio=True, nominal_voltage=voltage_config['nominal_voltage'])

    # Initialize dss upgrades file
    dss_commands_list = ["//This file has all the voltage upgrades\n"]
    upgrade_status = ''  # status - whether voltage upgrades done or not
    deciding_field = "deviation_severity"
    
    # determine voltage violations based on initial limits
    voltage_upper_limit = voltage_config["initial_upper_limit"]
    voltage_lower_limit = voltage_config["initial_lower_limit"]

    initial_xfmr_loading_df = get_thermal_equipment_info(compute_loading=True, upper_limit=thermal_config["transformer_upper_limit"], 
                                                        equipment_type="transformer", **simulation_params)
    initial_line_loading_df = get_thermal_equipment_info(compute_loading=True, upper_limit=thermal_config["line_upper_limit"], 
                                                        equipment_type="line", ignore_switch=ignore_switch, **simulation_params)
    initial_bus_voltages_df, initial_undervoltage_bus_list, initial_overvoltage_bus_list, \
        initial_buses_with_violations = get_bus_voltages(
            voltage_upper_limit=thermal_config['voltage_upper_limit'], voltage_lower_limit=thermal_config['voltage_lower_limit'],
            **simulation_params)

    initial_overloaded_xfmr_list = list(initial_xfmr_loading_df.loc[initial_xfmr_loading_df['status'] ==
                                                                    'overloaded']['name'].unique())
    initial_overloaded_line_list = list(initial_line_loading_df.loc[initial_line_loading_df['status'] ==
                                                                    'overloaded']['name'].unique())

    if os.path.exists(feeder_stats_json_file):
        feeder_stats = load_data(feeder_stats_json_file)
    else:
        feeder_stats = {"stage_results": []}
    feeder_stats["stage_results"].append( get_upgrade_stage_stats(dss, upgrade_stage="initial", upgrade_type="voltage", xfmr_loading_df=initial_xfmr_loading_df, line_loading_df=initial_line_loading_df, 
                                        bus_voltages_df=initial_bus_voltages_df, regcontrols_df=orig_regcontrols_df, capacitors_df=orig_capacitors_df) )
    dump_data(feeder_stats, feeder_stats_json_file, indent=2)
    scenario = get_scenario_name(enable_pydss_solve, pydss_volt_var_model)
    initial_results = UpgradeViolationResultModel(
        name = job_name, 
        scenario = scenario,
        stage = "initial",
        upgrade_type = "voltage",
        simulation_time_s = 0.0,
        thermal_violations_present = (len(initial_overloaded_xfmr_list) + len(initial_overloaded_line_list)) > 0,
        voltage_violations_present = (len(initial_undervoltage_bus_list) + len(initial_overvoltage_bus_list)) > 0,
        max_bus_voltage = initial_bus_voltages_df['max_per_unit_voltage'].max(),
        min_bus_voltage = initial_bus_voltages_df['min_per_unit_voltage'].min(),
        num_voltage_violation_buses = len(initial_undervoltage_bus_list) + len(initial_overvoltage_bus_list),
        num_overvoltage_violation_buses = len(initial_overvoltage_bus_list),
        voltage_upper_limit = voltage_upper_limit,
        num_undervoltage_violation_buses = len(initial_undervoltage_bus_list),
        voltage_lower_limit = voltage_lower_limit,
        max_line_loading = initial_line_loading_df['max_per_unit_loading'].max(),
        max_transformer_loading = initial_xfmr_loading_df['max_per_unit_loading'].max(),
        num_line_violations = len(initial_overloaded_line_list),
        line_upper_limit = thermal_config['line_upper_limit'],
        num_transformer_violations = len(initial_overloaded_xfmr_list),
        transformer_upper_limit = thermal_config['transformer_upper_limit'] 
    )
    temp_results = convert_dict_nan_to_none(dict(initial_results))
    if os.path.exists(overall_output_summary_filepath):
        overall_outputs = load_data(overall_output_summary_filepath)
        overall_outputs["violation_summary"].append(temp_results)
    else:
        overall_outputs = {"violation_summary": [temp_results]}
    dump_data(overall_outputs, overall_output_summary_filepath, indent=2, allow_nan=False)
    circuit_source = orig_ckt_info["source_bus"]
    bus_voltages_df, undervoltage_bus_list, overvoltage_bus_list, buses_with_violations = get_bus_voltages(
                voltage_upper_limit=voltage_upper_limit, voltage_lower_limit=voltage_lower_limit, **simulation_params)    
    logger.info(f"Number of overvoltage violations: {len(overvoltage_bus_list)}")
    logger.info(f"Number of undervoltage violations: {len(undervoltage_bus_list)}")
    # if there are no buses with violations based on initial check, don't get into upgrade process
    # directly go to end of file
    if len(buses_with_violations) <= 0:
        logger.info("Voltage Upgrades not Required.")
        upgrade_status = 'Voltage Upgrades not Required'  # status - whether voltage upgrades done or not
    # else, if there are bus violations based on initial check, start voltage upgrades process
    else:
        if create_plots:
            plot_voltage_violations(fig_folder=voltage_upgrades_directory, title="Bus violations before voltage upgrades_"+str(len(buses_with_violations)), 
                                    buses_with_violations=buses_with_violations, circuit_source=circuit_source, enable_detailed=True)
        # change voltage checking thresholds. determine violations based on final limits
        voltage_upper_limit = voltage_config["final_upper_limit"]
        voltage_lower_limit = voltage_config["final_lower_limit"]
        upgrade_status = 'Voltage Upgrades Required'  # status - whether voltage upgrades done or not
        logger.info("Voltage Upgrades Required.")
        comparison_dict = {"original": compute_voltage_violation_severity(
            voltage_upper_limit=voltage_upper_limit, voltage_lower_limit=voltage_lower_limit, **simulation_params)}
        best_setting_so_far = "original"
        # start with capacitors
        if voltage_config["capacitor_action_flag"] and len(orig_capacitors_df) > 0:
            capacitor_dss_commands = determine_capacitor_upgrades(voltage_upper_limit, voltage_lower_limit, default_capacitor_settings, orig_capacitors_df, 
                                                                  voltage_config, deciding_field, fig_folder=os.path.join(voltage_upgrades_directory, "interim"), 
                                                                  create_plots=create_plots, circuit_source=circuit_source,**simulation_params)
           
            bus_voltages_df, undervoltage_bus_list, overvoltage_bus_list, buses_with_violations = get_bus_voltages(
                voltage_upper_limit=voltage_upper_limit, voltage_lower_limit=voltage_lower_limit, **simulation_params)   
            
            if (len(buses_with_violations) > 0):
                # if violations increased after capacitor modifications, remove the capacitor changes.
                comparison_dict["after_capacitor_modifications"] = compute_voltage_violation_severity(voltage_upper_limit=voltage_upper_limit, voltage_lower_limit=voltage_lower_limit, **simulation_params)
                if comparison_dict["after_capacitor_modifications"][deciding_field] < comparison_dict[best_setting_so_far][deciding_field]:
                    best_setting_so_far = "after_capacitor_modifications"
                    dss_commands_list = dss_commands_list + capacitor_dss_commands
                else:
                    reload_dss_circuit(dss_file_list=initial_dss_file_list, commands_list=dss_commands_list, **simulation_params)
                    bus_voltages_df, undervoltage_bus_list, overvoltage_bus_list, buses_with_violations = get_bus_voltages(
                                    voltage_upper_limit=voltage_upper_limit, voltage_lower_limit=voltage_lower_limit, **simulation_params)   
        else:
            logger.info("No capacitor banks exist in the system")
        # next: existing regulators
        # Do a settings sweep of existing reg control devices (other than sub LTC) after correcting their other parameters such as ratios etc
        if voltage_config["existing_regulator_sweep_action"] and (len(orig_regcontrols_df) > 0) and (len(buses_with_violations) > 0):
            # first correct regcontrol parameters (ptratio) including substation LTC, if present
            # then perform settings sweep and choose best setting.
            logger.info("Settings sweep for existing reg control devices (excluding substation LTC).")
            regcontrols_df, reg_sweep_commands_list = sweep_and_choose_regcontrol_setting(voltage_config=voltage_config, initial_regcontrols_df=orig_regcontrols_df, 
                                                        upper_limit=voltage_upper_limit, lower_limit=voltage_lower_limit, 
                                                        dss_file_list=initial_dss_file_list, deciding_field=deciding_field, correct_parameters=True, 
                                                        exclude_sub_ltc=True, only_sub_ltc=False, previous_dss_commands_list=dss_commands_list, 
                                                        fig_folder=os.path.join(voltage_upgrades_directory, "interim"), create_plots=create_plots, circuit_source=circuit_source,
                                                        title="Bus violations after existing vreg sweep",
                                                        **simulation_params)
            # added to commands list only if it is different from original
            dss_commands_list = dss_commands_list + reg_sweep_commands_list
            # determine voltage violations after changes
            bus_voltages_df, undervoltage_bus_list, overvoltage_bus_list, buses_with_violations = get_bus_voltages(voltage_upper_limit=voltage_upper_limit, voltage_lower_limit=voltage_lower_limit, **simulation_params)
        # Writing out the results before adding new devices
        logger.info("Write upgrades to dss file, before adding new devices.")
        write_text_file(string_list=dss_commands_list, text_file_path=voltage_upgrades_dss_filepath)
        # Use this block for adding a substation LTC, correcting its settings and running a sub LTC settings sweep.
        comparison_dict["before_addition_of_new_device"]= compute_voltage_violation_severity(
            voltage_upper_limit=voltage_upper_limit, voltage_lower_limit=voltage_lower_limit, **simulation_params)
        best_setting_so_far = "before_addition_of_new_device"
        if (voltage_config['use_ltc_placement']) and (len(buses_with_violations) > 0):
            subltc_results_dict = determine_substation_ltc_upgrades(voltage_upper_limit=voltage_upper_limit, voltage_lower_limit=voltage_lower_limit, 
                                    orig_regcontrols_df=orig_regcontrols_df, orig_ckt_info=orig_ckt_info, circuit_source=circuit_source, 
                                    default_subltc_settings=default_subltc_settings, voltage_config=voltage_config, dss_file_list=initial_dss_file_list, 
                                    comparison_dict=comparison_dict, deciding_field=deciding_field, previous_dss_commands_list=dss_commands_list, 
                                    best_setting_so_far=best_setting_so_far, fig_folder=os.path.join(voltage_upgrades_directory, "interim"), create_plots=create_plots, 
                                    default_capacitor_settings=default_capacitor_settings, **simulation_params)
            best_setting_so_far = subltc_results_dict["best_setting_so_far"]
            comparison_dict = subltc_results_dict["comparison_dict"]
            subltc_upgrade_commands = subltc_results_dict["subltc_upgrade_commands"]
            dss_commands_list = dss_commands_list + subltc_upgrade_commands
            # determine voltage violations after changes
            bus_voltages_df, undervoltage_bus_list, overvoltage_bus_list, buses_with_violations = get_bus_voltages(
                voltage_upper_limit=voltage_upper_limit, voltage_lower_limit=voltage_lower_limit, **simulation_params)

        if len(buses_with_violations) >= min((100 * len(initial_buses_with_violations)), 500, len(dss.Circuit.AllBusNames())):
            # if number of buses with violations is very high, the loop for adding new regulators will take very long
            # so disable this block
            logger.info(f"At this point, number of buses with violations is {len(buses_with_violations)}, but initial "
                        f"number of buses with violations is {len(initial_buses_with_violations)}")
            logger.info("So disable option for addition of new regulators")
            voltage_config["place_new_regulators"] = False
        if voltage_config["place_new_regulators"] and (len(buses_with_violations) > 0):
            new_reg_results_dict = determine_new_regulator_upgrades(voltage_config=voltage_config, buses_with_violations=buses_with_violations, 
                                             voltage_upper_limit=voltage_upper_limit, voltage_lower_limit=voltage_lower_limit, 
                                             deciding_field=deciding_field, circuit_source=circuit_source, 
                                             default_regcontrol_settings=default_regcontrol_settings, comparison_dict=comparison_dict, 
                                             best_setting_so_far=best_setting_so_far, dss_file_list=initial_dss_file_list, 
                                             previous_dss_commands_list=dss_commands_list, fig_folder=os.path.join(voltage_upgrades_directory, "interim"), 
                                             create_plots=create_plots, **simulation_params)
            best_setting_so_far = new_reg_results_dict["best_setting_so_far"]
            comparison_dict = new_reg_results_dict["comparison_dict"]
            new_reg_upgrade_commands = new_reg_results_dict["new_reg_upgrade_commands"]
            dss_commands_list = dss_commands_list + new_reg_upgrade_commands
            # determine voltage violations after changes
            bus_voltages_df, undervoltage_bus_list, overvoltage_bus_list, buses_with_violations = get_bus_voltages(
                voltage_upper_limit=voltage_upper_limit, voltage_lower_limit=voltage_lower_limit, **simulation_params)           

    dss_commands_list.append(f"Set MaxControlIter={simulation_params['max_control_iterations']}")
    if any("new " in string.lower() for string in dss_commands_list):  # if new equipment is added.
        dss_commands_list.append("CalcVoltageBases")
    dss_commands_list.append("Solve")
    write_text_file(string_list=dss_commands_list, text_file_path=voltage_upgrades_dss_filepath)
    redirect_command_list = create_upgraded_master_dss(dss_file_list=initial_dss_file_list + [voltage_upgrades_dss_filepath], upgraded_master_dss_filepath=upgraded_master_dss_filepath,
                                                       original_master_filename=os.path.basename(master_path))
    write_text_file(string_list=redirect_command_list, text_file_path=upgraded_master_dss_filepath)
    reload_dss_circuit(dss_file_list=[upgraded_master_dss_filepath], commands_list=None, **simulation_params,)
    # reading new objects (after upgrades)
    new_ckt_info = get_circuit_info()
    new_regcontrols_df = get_regcontrol_info(correct_PT_ratio=True, nominal_voltage=voltage_config["nominal_voltage"])
    new_capacitors_df = get_capacitor_info(correct_PT_ratio=True, nominal_voltage=voltage_config['nominal_voltage'])
    processed_cap = get_capacitor_upgrades(orig_capacitors_df=orig_capacitors_df, new_capacitors_df=new_capacitors_df)
    processed_reg = get_regulator_upgrades(orig_regcontrols_df=orig_regcontrols_df, new_regcontrols_df=new_regcontrols_df, 
                                           orig_xfmrs_df=orig_xfmrs_df, new_ckt_info=new_ckt_info)
    all_processed = processed_cap + processed_reg
    m = AllUpgradesTechnicalResultModel(voltage=all_processed)
    temp = { "voltage": m.dict(by_alias=True)["voltage"]}
    dump_data(temp, output_json_voltage_upgrades_filepath, indent=2)
    bus_voltages_df, undervoltage_bus_list, overvoltage_bus_list, buses_with_violations = get_bus_voltages(
        voltage_upper_limit=voltage_upper_limit, voltage_lower_limit=voltage_lower_limit, **simulation_params)
    
    xfmr_loading_df = get_thermal_equipment_info(compute_loading=True, upper_limit=thermal_config["transformer_upper_limit"], 
                                                equipment_type="transformer", **simulation_params)
    line_loading_df = get_thermal_equipment_info(compute_loading=True, upper_limit=thermal_config["line_upper_limit"], 
                                                equipment_type="line", ignore_switch=ignore_switch, **simulation_params)
    overloaded_xfmr_list = list(xfmr_loading_df.loc[xfmr_loading_df['status'] == 'overloaded']['name'].unique())
    overloaded_line_list = list(line_loading_df.loc[line_loading_df['status'] == 'overloaded']['name'].unique())
    if (upgrade_status == "Voltage Upgrades Required") and create_plots:
        plot_voltage_violations(fig_folder=voltage_upgrades_directory, title="Bus violations after voltage upgrades_"+str(len(buses_with_violations)), 
                                    buses_with_violations=buses_with_violations, circuit_source=circuit_source, enable_detailed=True)
    if os.path.exists(feeder_stats_json_file):
        feeder_stats = load_data(feeder_stats_json_file)
    else:
        feeder_stats = {}
    feeder_stats["stage_results"].append( get_upgrade_stage_stats(dss, upgrade_stage="final", upgrade_type="voltage", xfmr_loading_df=xfmr_loading_df, line_loading_df=line_loading_df, 
                                        bus_voltages_df=bus_voltages_df, regcontrols_df=new_regcontrols_df, capacitors_df=new_capacitors_df) )
    dump_data(feeder_stats, feeder_stats_json_file, indent=2) 
    end_time = time.time()
    logger.info(f"Simulation end time: {end_time}")
    simulation_time = end_time - start_time
    logger.info(f"Simulation time: {simulation_time}")
    final_results = UpgradeViolationResultModel(
        name = job_name, 
        scenario = scenario,
        stage = "final",
        upgrade_type = "voltage",
        simulation_time_s = simulation_time,
        thermal_violations_present = (len(overloaded_xfmr_list) + len(overloaded_line_list)) > 0,
        voltage_violations_present = (len(undervoltage_bus_list) + len(overvoltage_bus_list)) > 0,
        max_bus_voltage = bus_voltages_df['max_per_unit_voltage'].max(),
        min_bus_voltage = bus_voltages_df['min_per_unit_voltage'].min(),
        num_voltage_violation_buses = len(undervoltage_bus_list) + len(overvoltage_bus_list),
        num_overvoltage_violation_buses = len(overvoltage_bus_list),
        voltage_upper_limit = voltage_upper_limit,
        num_undervoltage_violation_buses = len(undervoltage_bus_list),
        voltage_lower_limit = voltage_lower_limit,
        max_line_loading = line_loading_df['max_per_unit_loading'].max(),
        max_transformer_loading = xfmr_loading_df['max_per_unit_loading'].max(),
        num_line_violations = len(overloaded_line_list),
        line_upper_limit = thermal_config['line_upper_limit'],
        num_transformer_violations = len(overloaded_xfmr_list),
        transformer_upper_limit = thermal_config['transformer_upper_limit']
    )
    temp_results = convert_dict_nan_to_none(dict(final_results))
    if os.path.exists(overall_output_summary_filepath):
        overall_outputs = load_data(overall_output_summary_filepath)
        overall_outputs["violation_summary"].append(temp_results)
    else:
        overall_outputs = {"violation_summary": [temp_results]}
    dump_data(overall_outputs, overall_output_summary_filepath, indent=2, allow_nan=False)
    


import json
import logging
import os
import pathlib
import sqlite3
from abc import ABC, abstractmethod

from disco.storage.db import (
    Task,
    Scenario,
    Job,
    Report,
    FeederHead,
    FeederLosses,
    Metadata,
    ThermalMetrics,
    VoltageMetrics,
    SnapshotTimePoints,
    HostingCapacity,
    PvDistances,
)
from disco.storage.exceptions import IngestionError

logger = logging.getLogger(__name__)


class IngesterBase(ABC):
    """Abstract ingester class"""
    
    data_class = None
    
    def __init__(self, database):
        self.database = database

    @abstractmethod
    def ingest(self, data):
        """Ingest data into database via sqlalchemy"""
    
    def _perform_ingestion(self, columns, data):
        """
        Parameters
        ----------
        columns: list[str]
        data: list[dict]
        """
        table = self.data_class.__table__.name
        values = ", ".join(["?"] * len(columns))
        columns = ", ".join(columns)
        conn = sqlite3.connect(self.database)
        try:
            sql = f"INSERT INTO {table} ({columns}) VALUES ({values})"
            conn.executemany(sql, data)
            conn.commit()
        finally:
            conn.close()
        logger.info("Ingestion success - '%s'.", table)


class TaskIngester(IngesterBase):
    """Class used for ingesting parsed task into database"""
    
    data_class = Task

    def ingest(self, task):
        """Ingest task data into task table in database"""
        columns = self.data_class.__table__.columns.keys()
        data = [tuple([task[column] for column in columns])]
        self._perform_ingestion(columns=columns, data=data)
        indexes = {task["name"]: task["id"]}
        return indexes


class ScenarioIngester(IngesterBase):
    """Class used for ingesting parsed scenarios into database"""
    
    data_class = Scenario
    
    def ingest(self, scenarios):
        """Ingest a list of scenarios into scenario table in database"""
        columns = self.data_class.__table__.columns.keys()
        data = [tuple([item[column] for column in columns]) for item in scenarios]
        self._perform_ingestion(columns=columns, data=data)
        indexes = {self._generate_identifier(item): item["id"] for item in scenarios}
        return indexes
    
    @staticmethod
    def _generate_identifier(item):
        fields = ["job_id", "simulation_type", "name"]
        values = [item[k] for k in fields]
        return " | ".join(values)


class JobIngester(IngesterBase):
    """Class used for ingesting parsed jobs into database"""
    
    data_class = Job
    
    def ingest(self, jobs):
        """Ingest a list of jobs into job table in database"""
        columns = self.data_class.__table__.columns.keys()
        data = [tuple([item[column] for column in columns]) for item in jobs]
        self._perform_ingestion(columns=columns, data=data)
        indexes = {job["name"]: job["id"] for job in jobs}
        return indexes


class ReportIngester(IngesterBase):
    """Class used for ingesting parsed reports into database"""
    
    data_class = Report

    def ingest(self, reports):
        """Ingest parsed reports into report table in database"""
        columns = self.data_class.__table__.columns.keys()
        data = [tuple([item[column] for column in columns]) for item in reports]
        self._perform_ingestion(columns=columns, data=data)
        indexes = {report["file_name"]: report["id"] for report in reports}
        return indexes


class TableIngesterMixin:
    """Mixin class for handling report table ingestion"""

    def ingest(self, objects):
        columns = self.data_class.__table__.columns.keys()
        data = [tuple([item[column] for column in columns]) for item in objects]
        self._perform_ingestion(columns=columns, data=data)
        if objects and "id" in objects[0]:
            indexes = {self._generate_identifier(item): item["id"] for item in objects}
        else:
            indexes = []
        return indexes

    @staticmethod
    def _generate_identifier(item):
        fields = [
            "name",
            "substation",
            "feeder",
            "placement",
            "sample",
            "penetration_level",
            "scenario"
        ]
        values = [str(item.get(k, None)) for k in fields]
        return " | ".join(values)


class FeederHeadIngester(TableIngesterMixin, IngesterBase):
    """Class used for ingesting feeder head data into database"""
    
    data_class = FeederHead


class FeederLossesIngester(TableIngesterMixin, IngesterBase):
    """Class used for ingesting feeder losses data into database"""
    
    data_class = FeederLosses


class MetadataIngester(TableIngesterMixin, IngesterBase):
    """Class used for ingesting metadata into database"""
    
    data_class = Metadata


class ThermalMetricsIngester(TableIngesterMixin, IngesterBase):
    """Class used for ingesting thermal metrics into database"""
    
    data_class = ThermalMetrics
    

class VoltageMetricsIngester(TableIngesterMixin, IngesterBase):
    """Class used for ingesting voltage metrics into database"""
    
    data_class = VoltageMetrics
    
    def _generate_identifier(self, item):
        """Further customize identifier by adding node_type"""
        identifier = super()._generate_identifier(item)
        identifier += f" | {item['node_type']}"
        return identifier


class SnapshotTimePointsIngester(TableIngesterMixin, IngesterBase):

    data_class = SnapshotTimePoints
    
    @staticmethod
    def _generate_identifier(item):
        fields = ["job_id", "name"]
        values = [item[k] for k in fields]
        return " | ".join(values)


class HostingCapacityIngester(IngesterBase):

    data_class = HostingCapacity
    
    @staticmethod
    def _generate_identifier(item):
        fields = ["hc_type", "metric_type", "feeder", "scenario", "time_point"]
        values = [str(item[k]) for k in fields]
        return " | ".join(values)

    def ingest(self, hc_results):
        """Ingest a list of hosting capacity results into hosting_capacity table in database"""
        columns = self.data_class.__table__.columns.keys()
        data = [tuple([item[column] for column in columns]) for item in hc_results]
        self._perform_ingestion(columns=columns, data=data)
        indexes = {self._generate_identifier(item): item["id"] for item in hc_results}
        return indexes

class PvDistancesIngester(TableIngesterMixin, IngesterBase):

    data_class = PvDistances


class OutputIngester(IngesterBase):
    """Class used for ingesting all parsed results into SQLite database"""

    def ingest(self, data):
        """Ingest DISCO results into SQLite3 database"""
        indexes = {}   
        indexes["task"] = self._ingest_task(data["task"])
        indexes["jobs"] = self._ingest_jobs(data["jobs"])
        indexes["scenarios"] = self._ingest_scenarios(data["scenarios"])
        indexes["reports"] = self._ingest_reports(data["reports"])
        indexes["feeder_head"] = self._ingest_feeder_head(data["feeder_head"])
        indexes["feeder_losses"] = self._ingest_feeder_losses(data["feeder_losses"])
        indexes["metadata"] = self._ingest_metadata(data["metadata"])
        indexes["thermal_metrics"] = self._ingest_thermal_metrics(data["thermal_metrics"])
        indexes["voltage_metrics"] = self._ingest_voltage_metrics(data["voltage_metrics"])
        if "snapshot_time_points" in data:
            res = self._ingest_snapshot_time_points(data["snapshot_time_points"])
            indexes["snapshot_time_points"] = res
        indexes["hosting_capacity"] = self._ingest_hosting_capacity(data["hosting_capacity"])
        if "pv_distances" in data:
            indexes["pv_distances"] = self._ingest_pv_distances(data["pv_distances"])
        else:
            logger.warning("Weighted-average PV distances are not present.")
        return indexes

    def _ingest_task(self, task):
        """Ingest task via TaskIngester
        
        Parameters
        ----------
        task: dict
        """
        ingester = TaskIngester(self.database)
        index = ingester.ingest(task)
        return index

    def _ingest_jobs(self, jobs):
        """Ingest jobs via JobIngester
        
        Parameters
        ----------
        jobs: list[dict]
        """
        ingester = JobIngester(self.database)
        indexes = ingester.ingest(jobs)
        return indexes
    
    def _ingest_scenarios(self, scenarios):
        """Ingest scenarios via ScenarioIngester
        
        Parameters
        ----------
        scenarios: list[dict]
        """
        ingester = ScenarioIngester(self.database)
        indexes = ingester.ingest(scenarios)
        return indexes

    def _ingest_reports(self, reports):
        """Ingest reports via ReportIngester

        Parameters
        ----------
        reports: dict(str=dict)
        """
        ingester = ReportIngester(self.database)
        indexes = ingester.ingest(reports)
        return indexes
    
    def _ingest_feeder_head(self, feeder_head):
        """Ingest feeder head via FeederHeadIngester
        
        Parameters
        ----------
        feeder_head: list[dict]
        """
        ingester = FeederHeadIngester(self.database)
        indexes = ingester.ingest(feeder_head)
        return indexes
    
    def _ingest_feeder_losses(self, feeder_losses):
        """Ingest feeder losses via FeederLossesIngester
        
        Parameters
        ----------
        feeder_losses: list[dict]
        """
        ingester = FeederLossesIngester(self.database)
        indexes = ingester.ingest(feeder_losses)
        return indexes
    
    def _ingest_metadata(self, metadata):
        """Ingest metadata via MetadataIngester
        
        Parameters
        ----------
        metadata: list[dict]
        """
        ingester = MetadataIngester(self.database)
        indexes = ingester.ingest(metadata)
        return indexes
    
    def _ingest_thermal_metrics(self, thermal_metrics):
        """Ingest thermal metrics via ThermalMetricsIngester
        
        Parameters
        ----------
        thermal_metrics: list[dict]
        """
        ingester = ThermalMetricsIngester(self.database)
        indexes = ingester.ingest(thermal_metrics)
        return indexes
    
    def _ingest_voltage_metrics(self, voltage_metrics):
        """Ingest voltage metrics via VoltageMetricsIngester
        
        Parameters
        ----------
        voltage_metrics: list[dict]
        """
        ingester = VoltageMetricsIngester(self.database)
        indexes = ingester.ingest(voltage_metrics)
        return indexes
    
    def _ingest_snapshot_time_points(self, snapshot_time_points):
        """Ingest snapshot time points via SnapshotTimePointsIngester
        
        Parameters
        ----------
        snapshot_time_points: list[dict]
        """
        ingester = SnapshotTimePointsIngester(self.database)
        indexes = ingester.ingest(snapshot_time_points)
        return indexes
    
    def _ingest_hosting_capacity(self, hosting_capacity):
        """Ingest hosting capacity via HostingCapacityIngester
        
        Parameters
        ----------
        hosting_capacity: list[dict]
        """
        ingester = HostingCapacityIngester(self.database)
        indexes = ingester.ingest(hosting_capacity)
        return indexes

    def _ingest_pv_distances(self, pv_distances):
        """Ingest PV distances via PvDistancesIngester

        Parameters
        ----------
        pv_distances: list[dict]
        """
        ingester = PvDistancesIngester(self.database)
        indexes = ingester.ingest(pv_distances)
        return indexes


def dump_storage_index(output, indexes):
    """Dump storage indexes into a JSON file
    
    Parameters
    ----------
    output: str
    indexes: dict(str=dict)
    """
    filename = pathlib.Path(output) / "storage.json"
    with open(filename, "w") as f:
        json.dump(indexes, f, indent=2)
    return filename




import logging


logger = logging.getLogger(__name__)


def create_analysis_config(analysis_type, overrides=None):
    # dynamically get class from analysis module
    module = __import__("disco.analysis", fromlist=[analysis_type])
    analysis_class = getattr(module, analysis_type)
    analysis_config = analysis_class()

    if overrides:
        try:
            analysis_config.set_config_overrides(overrides)
        except ValueError:
            logging.exception("Failed to set config overrides")
            raise

    return analysis_config



"""CLI to automatically create an anlysis configuration."""

import logging
import sys

import click

from jade.jobs.job_post_process import JobPostProcess
from jade.loggers import setup_logging
from jade.utils.custom_click_options import CustomOptions
from jade.utils.utils import load_data
from disco.enums import ANALYSIS_MODEL_TYPES
from disco.analysis import load_custom_overrides
from disco.analysis.configure import create_analysis_config


@click.option(
    "-t", "--analysis-type",
    cls=CustomOptions,
    required=True,
    allowed_values=ANALYSIS_MODEL_TYPES,
    not_required_if='config_file',
    help="Analysis type"
)
@click.option(
    "-f", "--config-file",
    is_eager=True,
    type=click.Path(exists=True),
    help="config file containing overrides specific to analysis type"
)
@click.option(
    "-o", "--overrides",
    is_eager=True,
    callback=load_custom_overrides,
    multiple=True,
    default=[],
    help="Set individual overrides for each analysis input."
)
@click.option(
    "-l", "--list-defaults",
    is_flag=True,
    default=False,
    help="Show default options for analysis type"
)
@click.option(
    "--output-file",
    default="analysis-post-process-config.toml",
    help="Analysis output file name"
)
@click.option(
    "-v", "--verbose",
    is_flag=True,
    default=False,
    help="Enable verbose log output."
)
@click.command()
def generate_analysis(analysis_type, config_file, list_defaults,
                      overrides, output_file, verbose):
    """Generate Analysis configuration for job post-processing."""
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging("generate_analysis", None, console_level=level, packages=["disco"])

    if config_file is None:
        config = {}
    else:
        config = load_data(config)

    # config file overrides
    config_overrides = {}
    if 'overrides' in config:
        for override in config['overrides']:
            config_overrides[override] = config['overrides'][override]

    # explicit user overrides
    if overrides:
        config_overrides.update(overrides)

    analysis_config = create_analysis_config(analysis_type, config_overrides)

    if list_defaults:
        analysis_config.print_inputs()
        sys.exit()

    post_process = JobPostProcess("disco.analysis", analysis_type, data=config_overrides)
    post_process.dump_config(output_file)



"""Utility functions for the disco package."""
import json
import enum
import numpy as np
from pathlib import PosixPath, WindowsPath


class ExtendedJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, enum.Enum):
            return obj.value

        if isinstance(obj, PosixPath) or isinstance(obj, WindowsPath):
            return str(obj)

        if isinstance(obj, np.int32):
            return int(obj)
        
        if isinstance(obj, np.int64):
            return int(obj)
        
        return json.JSONEncoder.default(self, obj)
    


import click
from disco.postprocess.hosting_capacity_from_pydss import compute_hosting_capacity_for_pydss
from sqlalchemy import over


@click.command()
@click.option(
    "-j",
    "--jade-output-dir",
    help="Ouput directory after running jade runs",
)
@click.option(
    "-o",
    "--output-json",
    default="./hosting-capacity-from-pydss-results.json",
    show_default=True,
    help="Ouput directory after running jade runs",
)
@click.option(
    "-v",
    "--overvoltage-threshold",
    default=1.05,
    show_default=True,
    help="Over voltage threshold value default is 1.05 pu",
)
@click.option(
    "-t",
    "--thermal-threshold",
    default=100.0,
    show_default=True,
    help="Thermal overloading threshold default is 100.0",
)
def hosting_capacity_by_timestep(
    jade_output_dir, output_json, overvoltage_threshold, thermal_threshold
):

    """Parse all the jobs in JADE output directory and compute hosting capacity for each
    feeders by scenario and by time step taking pydss results. This script assumes that you have
    already ran time series simulation for all your scenarios and exported bus voltage and thermal
    overloading metrics. Hosting capacity is computed by analyzing maximum voltage and/or maximum thermal
    overloading across all samples and placements for given substation scenario combination."""
    compute_hosting_capacity_for_pydss(
        jade_output_dir, output_json, overvoltage_threshold, thermal_threshold
    )



"""Defines interface for model input data."""

import abc


class ModelInputDataInterface(abc.ABC):
    """Defines interface for model input data."""

    @abc.abstractmethod
    def generate_output_data(self, output_dir, include_pv_systems):
        """Generate output data.

        Parameters
        ----------
        output_dir : str
            Directory in which to generate output data.
        include_pv_systems : bool
            Whether to include PVSystems.dss in the simulation.

        """



"""Generates OpenDSS data for distribution simulations."""

from collections import namedtuple
import fileinput
import itertools
import logging
import os
import pathlib
import re
import shutil

from jade.exceptions import InvalidParameter
from jade.utils.timing_utils import timed_info
from jade.utils.utils import dump_data, modify_file, get_filenames_by_ext, \
    ExtendedJSONEncoder
from disco.enums import SimulationType
from disco.models.base import OpenDssDeploymentModel, PyDSSControllerModel, \
    ImpactAnalysisBaseModel
from disco.sources.base import SOURCE_CONFIGURATION_FILENAME
from disco.sources.gem.model_input_interface import ModelInputDataInterface


logger = logging.getLogger(__name__)


class OpenDssGenerator(ModelInputDataInterface):
    """Generates OpenDSS data for distribution simulations."""

    _loadshapes_file = "LoadShapes.dss"

    def __init__(self, data):
        self._feeders = []
        self._include_voltage_deviation = data.get('include_voltage_deviation', False)
        for feeder in data["feeders"]:
            if feeder["name"] == "":
                logger.warning("Feeder name is empty; skipping %s", feeder)
            elif "DEAD" in feeder["name"].upper():
                logger.warning("Skipping 'dead' feeder %s", feeder["name"])
            else:
                self._feeders.append(Feeder(feeder))

        feeder_dirnames = set()
        for feeder in self._feeders:
            feeder_dirname = self._feeder_dirname(feeder)
            if feeder_dirname in feeder_dirnames:
                raise InvalidParameter(f"duplicate feeder dirname: {feeder_dirname}")
            feeder_dirnames.add(feeder_dirname)

            deployment_names = set()
            for deployment in feeder.deployments:
                if deployment.name in deployment_names:
                    raise InvalidParameter(
                        f"duplicate deployment found: {deployment.name}"
                    )
                deployment_names.add(deployment.name)

        self._modified_master_files = set()
        self._output_dir = None

    @staticmethod
    def _create_name(inputs):
        return "__".join([str(x) for x in [inputs.feeder.name,
                                           inputs.feeder.tag,
                                           inputs.deployment.name,
                                           inputs.dc_ac_ratio,
                                           inputs.kva_to_kw_rating]])

    @staticmethod
    def _feeder_dirname(feeder):
        return feeder.name + "__" + feeder.tag

    def _feeder_dir(self, output_dir, feeder):
        feeder_dirname = self._feeder_dirname(feeder)
        return os.path.join(output_dir, feeder_dirname)

    @staticmethod
    def _opendss_dir(feeder_dir):
        return os.path.join(feeder_dir, "OpenDSS")

    @staticmethod
    def _loadshapes_dir(feeder_dir):
        return os.path.join(feeder_dir, "LoadShapes")

    @staticmethod
    def _deployment_dir(feeder_dir):
        return os.path.join(feeder_dir, "PVDeployments")

    @staticmethod
    def _deployment_file(deployment_dir, deployment):
        return os.path.join(deployment_dir, deployment.name + ".dss")

    @staticmethod
    def _master_file(feeder_dir):
        return os.path.join(OpenDssGenerator._opendss_dir(feeder_dir),
                            "Master.dss")

    @staticmethod
    def _new_master_file(feeder_dir):
        return os.path.join(OpenDssGenerator._opendss_dir(feeder_dir),
                            "MasterDisco.dss")

    def _configurations_dir(self):
        return os.path.join(self._output_dir, "configurations")

    @staticmethod
    def _get_pattern(data, pattern_type):
        assert pattern_type in ("dc_ac_ratio_patterns", "kva_to_kw_rating_patterns")
        patterns = set()
        for item in data[pattern_type]:
            val = None
            string = None
            if "all" in item:
                val = item["all"]
                string = "all_" + str(val)
            pattern = (val, string)
            if pattern in patterns:
                raise InvalidParameter("duplicate {pattern_type} {pattern}")
            patterns.add(pattern)

        return list(patterns)

    @timed_info
    def generate_output_data(self, output_dir, include_pv_systems, simulation_model):
        self._output_dir = output_dir

        jobs = []
        for feeder in self._feeders:
            if not self._check_required_feeder_files(feeder):
                continue
            feeder_dir = self._feeder_dir(self._output_dir, feeder)
            opendss_dir = self._opendss_dir(feeder_dir)
            self._copy_dss_files(feeder.opendss_location, opendss_dir)
            for deployment in feeder.deployments:
                self._copy_deployment_data(deployment, feeder_dir,
                                           include_pv_systems)
                for args in itertools.product(deployment.dc_ac_ratios,
                                              deployment.kva_to_kw_ratings):
                    inputs = DeploymentInputs(feeder, deployment, *args)
                    job = self._generate_job(inputs, output_dir, feeder_dir, simulation_model)
                    jobs.append(job)

        if jobs:
            self._create_config_file(jobs)

        logger.info("Done generating data in %s", output_dir)

    def _create_config_file(self, jobs):
        config_file = os.path.join(self._output_dir, SOURCE_CONFIGURATION_FILENAME)
        dump_data(jobs, config_file, indent=2, cls=ExtendedJSONEncoder)

    @staticmethod
    def _check_required_feeder_files(feeder):
        is_valid = True

        if not os.path.exists(feeder.opendss_location):
            logger.error(
                "Possible bad feeder %s - tag %s. Path isn't present: %s",
                feeder.name, feeder.tag, feeder.opendss_location
            )
            is_valid = False

        required = ["Master.dss"]
        for filename in required:
            path = os.path.join(feeder.opendss_location, filename)
            if not os.path.exists(path):
                logger.error(
                    "Feeder-tag=%s-%s %s does not exist in %s",
                    feeder.name, feeder.tag, filename, feeder.opendss_location
                )
                is_valid = False

        return is_valid

    @staticmethod
    def _copy_dss_files(src, dst):
        """Copy .dss files recursively, fixing data-file paths as necessary.

        Parameters
        ----------
        src : str
            source directory
        dst : str
            destination directory

        """
        os.makedirs(dst, exist_ok=True)
        target_dirs = set([dst])

        for path in get_filenames_by_ext(src, ".dss"):
            relpath = os.path.relpath(path, src)
            target_path = os.path.join(dst, relpath)
            dirname = os.path.dirname(target_path)
            # This tracks directory creations in memory in order to reduce
            # fileystem operations which could be slow on distributed systems
            # like Lustre.
            if dirname not in target_dirs:
                os.mkdir(dirname)
                target_dirs.add(dirname)
            shutil.copyfile(path, target_path)
            logger.debug("Copied %s to %s", path, target_path)

            src_dir = os.path.abspath(os.path.dirname(path))
            # TODO: add support for copying data files if we use this source model again.
            OpenDssGenerator.make_data_file_references_absolute(src_dir, target_path)

    @staticmethod
    def make_data_file_references_absolute(src_dir, filename):
        """Change the path to any data file referenced in a .dss file to its
        absolute path."""
        # Example line:
        # New Loadshape.Residential1234 npts=123456 minterval=5 mult=[file=../BuildingData/Dataset_12_34/Residential/RES1234/LoadProfiles/12345.csv, col=1, header=no]
        regex = re.compile(r"file=(.*), col=")
        def replace_func(match):
            path = os.path.normpath(match.group(1).replace("\\", "/"))
            new_path = os.path.normpath(os.path.join(src_dir, path))
            return "file=" + new_path + ", col="

        with fileinput.input(files=[filename], inplace=True) as f_in:
            for line in f_in:
                line = re.sub(regex, replace_func, line)
                print(line, end="")
                logger.debug("line=%s", line)

    def _copy_deployment_data(self, deployment, feeder_dir, include_pv_systems):
        self._attach_pv_deployment(deployment, feeder_dir)

        new_master_file = self._new_master_file(feeder_dir)
        if new_master_file not in self._modified_master_files:
            path_master_file = self._master_file(feeder_dir)
            # Solve will be called from the PV deployment file.
            regex_solve = re.compile(r"^\s*Solve\s*$")
            modify_file(path_master_file, comment_out_solve, regex_solve)
            if not include_pv_systems:
                # Comment out any PVSystems file in the original data. The QSTS
                # code reads from the PVDeployments folder instead. The new
                # masterfile is named "MasterDisco.dss"
                modify_file(path_master_file, comment_out_pv_systems,
                            "PVSystems.dss")
                logger.debug("Comment out existing PVSystems file. %s",
                             path_master_file)

            shutil.move(path_master_file, new_master_file)
            logger.debug("Moved %s to %s", path_master_file, new_master_file)
            self._modified_master_files.add(new_master_file)

    def _attach_pv_deployment(self, deployment, feeder_dir):
        """Add redirects to the master file file within the PV deployment."""
        deployment_dir = self._deployment_dir(feeder_dir)
        os.makedirs(deployment_dir, exist_ok=True)

        # Copy the PV file to the deployments folder with deployment.name as
        # the file name.
        pv_file = self._deployment_file(deployment_dir, deployment)
        if os.path.exists(pv_file):
            return

        master_file = os.path.join(feeder_dir, "OpenDSS", "MasterDisco.dss")

        # Below Solve is added to the file because OpenDSS will not properly
        # initialized internal memory structures unless Solve is called after
        # all new elements have been added.

        if not deployment.pv_locations:
            # TODO: hack
            with open(pv_file, "w") as f_out:
                f_out.write(f"Redirect {master_file}\n")
                f_out.write("\nSolve\n")
            return

        pv_locations = deployment.pv_locations
        if not isinstance(pv_locations, list):
            pv_locations = [pv_locations]

        with fileinput.input(pv_locations) as fr, open(pv_file, "w") as fw:
            for line in fr:
                fw.write(line)

        logger.debug("Copied %s to %s", deployment.name, pv_file)

        # Include MasterDisco.dss within this file.
        tmp = pv_file + ".tmp"
        with open(tmp, "w") as fp_out:
            fp_out.write(f"Redirect {master_file}\n\n")
            with open(pv_file) as fp_in:
                for line in fp_in:
                    fp_out.write(line)
            fp_out.write("\nSolve\n")
        shutil.move(tmp, pv_file)

    def _generate_job(self, inputs, output, feeder_dir, simulation_model):
        deployment_dir = self._deployment_dir(feeder_dir)
        os.makedirs(deployment_dir, exist_ok=True)
        feeder = inputs.feeder
        deployment = inputs.deployment
        opendss_dir = os.path.join(feeder_dir, "OpenDSS")

        project_data = {
            "penetration": deployment.penetration,
            "placement_type": deployment.placement_type,
            "sample": deployment.sample,
        }
        project_data.update(deployment.project_data)
        simulation_params = {
            "simulation_type": feeder.simulation_type.value,
            "start_time": feeder.start_time,
            "end_time": feeder.end_time,
            "step_resolution": feeder.step_resolution,
        }
        feeder_dirname = self._feeder_dirname(feeder)
        deployment_file = os.path.join(output, feeder_dirname, "PVDeployments", deployment.name + ".dss")
        if deployment.pydss_controllers:
            pydss_controllers = [
                PyDSSControllerModel(
                    controller_type=x["controller_type"],
                    name=x["name"],
                )
                for x in deployment.pydss_controllers
            ]
        else:
            pydss_controllers = None

        deployment_model = OpenDssDeploymentModel.validate(
            dict(
                deployment_file=deployment_file,
                feeder=feeder.name,
                dc_ac_ratio=inputs.dc_ac_ratio,
                kva_to_kw_rating=inputs.kva_to_kw_rating,
                directory=output,
                project_data=project_data,
                pydss_controllers=pydss_controllers,
            )
        )

        name = "__".join([
            feeder_dirname,
            str(inputs.dc_ac_ratio),
            str(inputs.kva_to_kw_rating),
            deployment.name,
            str(deployment.job_order)
        ])

        data = {
            "deployment": deployment_model,
            "simulation": simulation_params,
            "name": name,
            "model_type": simulation_model.__name__,
            "job_order": deployment.job_order
        }
        if issubclass(simulation_model, ImpactAnalysisBaseModel):
            data["base_case"] = inputs.feeder.base_case

        return simulation_model.validate(data).dict()


class Feeder:
    """Defines a feeder."""
    def __init__(self, data):
        self.name = data["name"]
        self.tag = data.get("tag", None)
        self.base_case = data.get("base_case", None)
        self.deployments = [Deployment(x) for x in data["deployments"]]
        self.opendss_location = data["opendss_location"]
        self.loadshape_location = data["loadshape_location"]
        self.simulation_type = SimulationType(data.get("simulation_type", "Snapshot"))
        self.start_time = data["start_time"]
        self.end_time = data["end_time"]
        self.step_resolution = data["step_resolution"]

        if self.simulation_type == SimulationType.SNAPSHOT:
            if self.start_time != self.end_time:
                raise InvalidParameter(
                    "start_time and end_time must be the same for snapshot "
                    f"simulations: {data}"
                )

        if not self.tag:
            self.tag = self.create_feeder_tag()

    def create_feeder_tag(self):
        """A TEMP solution of creating feeder tag from opendss_location.

        Suppose the directory structure is like this::

            opendss_location = (
                "/project/xxx/run2/distribution/xxx/feeder_models/"
                "opendss/xxx/scenarios/xxx/"
                "2050/2010-07-10_15-00-00-000/X/456/xxx"
            )

        The tag would become "2050__2010-07-10_15-00-00-000__X__456"

        """
        path = pathlib.Path(self.opendss_location)
        parents = [p.name for p in path.parents if p.name]
        if len(parents) > 4:
            tag = "__".join(reversed(parents[:4]))
        else:
            tag = "__".join(reversed(parents))
        return tag


class Deployment:
    """Defines a deployment."""
    def __init__(self, data):
        self.name = data["name"]
        self.dc_ac_ratios = data["dc_ac_ratios"]
        self.kva_to_kw_ratings = data["kva_to_kw_ratings"]
        self.loadshape_file = data["loadshape_file"]
        self.loadshape_location = data["loadshape_location"]
        self.job_order = data.get("job_order", None)

        # Handle legacy formats.
        if "pv_location" in data:
            self.pv_locations = [data["pv_location"]]
        else:
            self.pv_locations = data.get("pv_locations", [])
        self.pydss_controllers = data.get("pydss_controllers", [])
        if isinstance(self.pydss_controllers, dict):
            self.pydss_controllers = [self.pydss_controllers]

        for param in ("dc_ac_ratios", "kva_to_kw_ratings"):
            val = getattr(self, param)
            if val is None or val == []:
                setattr(self, param, [None])

        # Optional
        self.penetration = data.get("penetration", None)
        self.sample = data.get("sample", None)
        self.placement_type = data.get("placement_type", None)
        self.project_data = data.get("project_data", {})


DeploymentInputs = namedtuple(
    "Inputs",
    "feeder, deployment, dc_ac_ratio, kva_to_kw_rating"
)


def comment_out_pv_systems(line, *args, **_):
    """Comments out the redirect to a given string."""
    return line.replace("Redirect " + args[0], "!Redirect " + args[0])


def comment_out_solve(line, *args, **_):
    """Comments out the Solve command."""
    regex = args[0]
    match = regex.search(line)
    if match:
        return "!Solve"
    return line
